"""
Copyright 2023 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

import datetime as dt
from unittest.mock import patch

from pandas.testing import assert_frame_equal

from gs_quant.api.api_cache import InMemoryApiRequestCache, CacheEvent
from gs_quant.api.gs.data import GsDataApi, QueryType
from gs_quant.data import Dataset, DataContext


class NotExpectedToBeCalledSession:
    redirect_to_mds = True

    @classmethod
    def _get(cls, url, **kwargs):
        if url == "/data/datasets/FXSPOT_STANDARD":
            return {"id": "FXSPOT_STANDARD"}
        else:
            raise Exception("Not expecting to be called at this point")

    @classmethod
    def _post(cls, url, **kwargs):
        raise Exception("Not expecting to be called at this point")


class MarketDataErrorSession:
    redirect_to_mds = True

    @classmethod
    def _post(cls, url, **kwargs):
        return {
            "requestId": "890",
            "responses": [{"queryResponse": [{"errorMessages": ["Test Failure"]}]}],
        }


class FakeSession:
    redirect_to_mds = True

    @classmethod
    def _get(cls, url, **kwargs):
        if url == "/data/catalog/FXSPOT_STANDARD":
            return {
                "id": "FXSPOT_STANDARD",
                "fields": {
                    "date": {"type": "string", "format": "date"},
                    "assetId": {"type": "string"},
                    "spot": {
                        "type": "number",
                    },
                    "updateTime": {"type": "string", "format": "date-time"},
                },
            }
        elif url == "/data/datasets/FXSPOT_STANDARD":
            return {"id": "FXSPOT_STANDARD"}
        else:
            raise Exception("Need to mock _get request here")

    @classmethod
    def _post(cls, url, **kwargs):
        if url == "/data/FXSPOT_STANDARD/last/query":
            return {
                "requestId": "1234",
                "data": [
                    {
                        "date": "2023-10-25",
                        "assetId": "MATGYV0J9MPX534Z",
                        "bbid": "USDJPY",
                        "spot": 150.123,
                        "updateTime": "2023-10-25T21:53:56Z",
                    }
                ],
            }
        elif url == "/data/FXSPOT_STANDARD/query":
            return {
                "requestId": "5678",
                "data": [
                    {
                        "date": "2023-10-26",
                        "assetId": "MATGYV0J9MPX534Z",
                        "bbid": "USDJPY",
                        "spot": 152.234,
                        "updateTime": "2023-10-26T21:53:56Z",
                    }
                ],
            }
        elif url == "/data/measures":
            return {
                "requestId": "890",
                "responses": [
                    {
                        "queryResponse": [
                            {
                                "measure": "Curve",
                                "dataSetIds": ["DATASET_FOO"],
                                "entityTypes": ["ASSET"],
                                "response": {
                                    "data": [
                                        {
                                            "date": "2023-04-11",
                                            "assetId": "MATGYV0J9MPX534Z",
                                            "pricingLocation": "HKG",
                                            "name": "USDJPY",
                                            "spot": 133.0,
                                        },
                                        {
                                            "date": "2023-04-11",
                                            "assetId": "MATGYV0J9MPX534Z",
                                            "pricingLocation": "LDN",
                                            "name": "USDJPY",
                                            "spot": 134.0,
                                        },
                                        {
                                            "date": "2023-04-11",
                                            "assetId": "MATGYV0J9MPX534Z",
                                            "pricingLocation": "NYC",
                                            "name": "USDJPY",
                                            "spot": 136.0,
                                        },
                                    ]
                                },
                            }
                        ]
                    }
                ],
            }


class TestDataApiCache:

    def setup_method(self, test_method):
        self.cache = InMemoryApiRequestCache()
        GsDataApi.set_api_request_cache(self.cache)

    def teardown_method(self, test_method):
        GsDataApi.set_api_request_cache(None)

    def test_last_data(self):
        ds = Dataset("FXSPOT_STANDARD")
        with patch.object(GsDataApi, "get_session", return_value=FakeSession()):
            df = ds.get_data_last(as_of=dt.date(2023, 10, 25), bbid="USDJPY")
        with patch.object(
            GsDataApi, "get_session", return_value=NotExpectedToBeCalledSession()
        ):
            df2 = ds.get_data_last(dt.date(2023, 10, 25), bbid="USDJPY")
        assert not df.empty
        assert_frame_equal(df, df2)
        cache_events = self.cache.get_events()
        assert len(cache_events) == 4
        assert cache_events[0][0] == CacheEvent.PUT
        assert cache_events[1][0] == CacheEvent.PUT
        assert cache_events[2][0] == CacheEvent.GET
        assert cache_events[3][0] == CacheEvent.GET

    def test_query_data(self):
        ds = Dataset("FXSPOT_STANDARD")
        with patch.object(GsDataApi, "get_session", return_value=FakeSession()):
            df = ds.get_data(
                dt.date(2023, 10, 26), dt.date(2023, 10, 26), bbid="USDJPY"
            )
        with patch.object(
            GsDataApi, "get_session", return_value=NotExpectedToBeCalledSession()
        ):
            df2 = ds.get_data(
                dt.date(2023, 10, 26), dt.date(2023, 10, 26), bbid="USDJPY"
            )

        assert_frame_equal(df, df2)
        cache_events = self.cache.get_events()
        assert len(cache_events) == 4
        assert cache_events[0][0] == CacheEvent.PUT
        assert cache_events[1][0] == CacheEvent.PUT
        assert cache_events[2][0] == CacheEvent.GET
        assert cache_events[3][0] == CacheEvent.GET

    def test_market_data(self):
        asset_id = "MATGYV0J9MPX534Z"
        with DataContext(dt.date(2023, 4, 11), dt.date(2023, 4, 11)):
            q = GsDataApi.build_market_data_query([asset_id], QueryType.SPOT)

        with patch.object(
            GsDataApi, "get_session", return_value=MarketDataErrorSession()
        ):
            try:
                df = GsDataApi.get_market_data(q)
            except Exception:
                pass
        cache_events = self.cache.get_events()
        assert len(cache_events) == 0
        with patch.object(GsDataApi, "get_session", return_value=FakeSession()):
            df = GsDataApi.get_market_data(q)
        with patch.object(
            GsDataApi, "get_session", return_value=NotExpectedToBeCalledSession()
        ):
            df2 = GsDataApi.get_market_data(q)

        assert_frame_equal(df, df2)
        cache_events = self.cache.get_events()
        assert len(cache_events) == 2
        assert cache_events[0][0] == CacheEvent.PUT
        assert cache_events[1][0] == CacheEvent.GET



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

from gs_quant.target.common import *  # noqa
from gs_quant.target.common import PayReceive as _PayReceive  # noqa
from gs_quant.target.common import RiskMeasure as __RiskMeasure  # noqa
from gs_quant.target.common import RiskMeasureType, AssetClass, RiskMeasureUnit  # noqa
from gs_quant.base import EnumBase, RiskMeasureParameter  # noqa

from enum import Enum
from typing import Union
import datetime as dt


class PositionType(Enum):
    """Position type enumeration

    Enumeration of different position types for a portfolio or index

    """

    OPEN = "open"  #: Open positions (corporate action adjusted)
    CLOSE = "close"  #: Close positions (reflect trading activity on the close)
    ANY = 'any'


class DateLimit(Enum):
    """ Datetime date constants """

    LOW_LIMIT = dt.date(1952, 1, 1)


class PayReceive(EnumBase, Enum):
    """Pay or receive fixed"""

    Pay = 'Pay'
    Receive = 'Rec'
    Straddle = 'Straddle'

    @classmethod
    def _missing_(cls, key):
        if isinstance(key, _PayReceive):
            key = key.value

        return cls.Receive if key.lower() in ('receive', 'receiver') else super()._missing_(key)


class RiskMeasure(__RiskMeasure):

    def __lt__(self, other):
        if self.name != other.name:
            return self.name < other.name
        elif self.parameters is not None:
            if other.parameters is None:
                return False
            if not isinstance(other.parameters, type(self.parameters)):
                return self.parameters.parameter_type < other.parameters.parameter_type
            else:
                return repr(self.parameters) < repr(other.parameters)
        elif other.parameters is not None:
            return True
        return False

    def __repr__(self):
        return self.name or self.measure_type.name

    @property
    def pricing_context(self):
        from gs_quant.markets import PricingContext
        return PricingContext.current


class ParameterisedRiskMeasure(RiskMeasure):
    def __init__(self, asset_class: Union[AssetClass, str] = None,
                 measure_type: Union[RiskMeasureType, str] = None, unit: Union[RiskMeasureUnit, str] = None,
                 value: Union[float, str] = None, parameters: RiskMeasureParameter = None, name: str = None):
        super().__init__(asset_class=asset_class, measure_type=measure_type, unit=unit, value=value, name=name)
        if parameters:
            if isinstance(parameters, RiskMeasureParameter):
                self.parameters = parameters
            else:
                raise TypeError(f"Unsupported parameter {parameters}")

    def __repr__(self):
        name = self.name or self.measure_type.name
        params = None
        if self.parameters:
            params = self.parameters.as_dict()
            params.pop('parameter_type', None)
            sorted_keys = sorted(params.keys(), key=lambda x: x.lower())
            params = ', '.join(
                [f'{k}:{params[k].value if isinstance(params[k], EnumBase) else params[k]}' for k in sorted_keys])
        return name + '(' + params + ')' if params else name

    def parameter_is_empty(self):
        return self.parameters is None



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import copy
import re
import warnings
from functools import reduce

import pandas as pd

from gs_quant.api.gs.backtests_xasset.response_datatypes.backtest_datatypes import TransactionCostConfig, \
    TradingCosts, FixedCostModel, ScaledCostModel, TransactionCostScalingType, AggregateCostModel, CostAggregationType
from gs_quant.backtests import actions as a
from gs_quant.backtests import triggers as t
from gs_quant.backtests.backtest_objects import TransactionModel, ConstantTransactionModel, ScaledTransactionModel, \
    AggregateTransactionModel

from gs_quant.backtests.strategy_systematic import StrategySystematic, DeltaHedgeParameters, TradeInMethod
from gs_quant.base import get_enum_value
from gs_quant.common import OptionType, BuySell
from gs_quant.instrument import EqOption, EqVarianceSwap
from gs_quant.markets.portfolio import Portfolio
from gs_quant.risk import EqDelta, EqSpot, EqGamma, EqVega
from gs_quant.target.backtests import BacktestSignalSeriesItem, BacktestTradingQuantityType, EquityMarketModel, \
    FlowVolBacktestMeasure
from gs_quant.common import TradeAs


def get_backtest_trading_quantity_type(scaling_type, risk):
    if scaling_type == a.ScalingActionType.size:
        return a.BacktestTradingQuantityType.quantity
    if scaling_type == a.ScalingActionType.NAV:
        return a.BacktestTradingQuantityType.NAV
    if risk == EqSpot:
        return BacktestTradingQuantityType.notional
    if risk == EqGamma:
        return BacktestTradingQuantityType.gamma
    if risk == EqVega:
        return BacktestTradingQuantityType.vega
    raise ValueError(f"unable to translate {scaling_type} and {risk}")


def is_synthetic_forward(priceable):
    is_portfolio = isinstance(priceable, Portfolio)
    is_syn_fwd = is_portfolio
    if is_portfolio:
        is_size_two = len(priceable) == 2
        is_syn_fwd = is_size_two
        if is_size_two:
            has_two_eq_options = isinstance(priceable[0], EqOption) and isinstance(priceable[1], EqOption)
            is_syn_fwd &= has_two_eq_options
            if has_two_eq_options:
                is_syn_fwd &= (priceable[0].underlier == priceable[1].underlier) and \
                              (priceable[0].expiration_date == priceable[1].expiration_date) and \
                              (priceable[0].strike_price == priceable[1].strike_price)
                is_syn_fwd &= (OptionType.Call, BuySell.Buy) and (OptionType.Put, BuySell.Sell) in {
                              (priceable[0].option_type, priceable[0].buy_sell),
                              (priceable[1].option_type, priceable[1].buy_sell)}

    return is_syn_fwd


class BacktestResult:
    def __init__(self, results):
        self._results = results

    def get_measure_series(self, measure: FlowVolBacktestMeasure):
        data = next(iter(r.timeseries for r in self._results.risks if r.name == measure.value), ())
        df = pd.DataFrame.from_records(data)
        if len(df) == 0:
            return df

        df['date'] = pd.to_datetime(df['date'])
        return df.set_index('date').value

    def get_portfolio_history(self):
        data = []
        for item in self._results.portfolio:
            positions = list(map(lambda x: dict({'date': item['date'], 'quantity': x['quantity']}, **x['instrument']),
                                 item['positions']))
            data = data + positions

        return pd.DataFrame(data)

    def get_trade_history(self):
        data = []
        for item in self._results.portfolio:
            for transaction in item['transactions']:
                trades = list(map(lambda x: dict(
                    {'date': item['date'], 'quantity': x['quantity'], 'transactionType': transaction['type'],
                     'price': x['price'], 'cost': transaction.get('cost') if len(transaction['trades']) == 1 else None},
                    **x['instrument']), transaction['trades']))
                data = data + trades

        return pd.DataFrame(data)


class EquityVolEngine(object):

    @classmethod
    def check_strategy(cls, strategy):
        check_results = []
        if len(strategy.initial_portfolio) > 0:
            check_results.append('Error: initial_portfolio must be empty or None')

        # Validate Triggers

        if len(strategy.triggers) > 3:
            check_results.append('Error: Maximum of 3 triggers')

        if not all(isinstance(x, (t.AggregateTrigger, t.PeriodicTrigger)) for x in strategy.triggers):
            check_results.append('Error: Only AggregateTrigger and PeriodTrigger supported')

        # aggregate triggers composed of a dated and portfolio trigger define a signal
        aggregate_triggers = [x for x in strategy.triggers if isinstance(x, t.AggregateTrigger)]
        for at in aggregate_triggers:
            if not len(at.trigger_requirements.triggers) == 2:
                check_results.append('Error: AggregateTrigger must be composed of 2 triggers')
            if not len([x for x in at.trigger_requirements.triggers if isinstance(x, t.DateTriggerRequirements)]) == 1:
                check_results.append('Error: AggregateTrigger must be contain 1 DateTrigger')
            portfolio_triggers = [x for x in at.trigger_requirements.triggers
                                  if isinstance(x, t.PortfolioTriggerRequirements)]
            if not len(portfolio_triggers) == 1:
                check_results.append('Error: AggregateTrigger must be contain 1 PortfolioTrigger')
            if not (portfolio_triggers[0].data_source == 'len' and
                    portfolio_triggers[0].trigger_level == 0):
                check_results.append(
                    'Error: PortfolioTrigger.trigger_requirements must have data_source = \'len\' '
                    'and trigger_level = 0')

        # Validate Actions

        all_actions = reduce(lambda acc, x: acc + x, (map(lambda x: x.actions, strategy.triggers)), [])

        if any(isinstance(x, a.ExitPositionAction) for x in all_actions):
            warnings.warn('ExitPositionAction will be deprecated soon, use ExitTradeAction.', DeprecationWarning, 2)
        if any(isinstance(x, a.EnterPositionQuantityScaledAction) for x in all_actions):
            warnings.warn('EnterPositionQuantityScaledAction will be deprecated soon, use AddScaledTradeAction.',
                          DeprecationWarning, 2)

        if not all(isinstance(x, (a.EnterPositionQuantityScaledAction, a.HedgeAction, a.ExitPositionAction,
                                  a.ExitTradeAction, a.AddTradeAction, a.AddScaledTradeAction)) for x in all_actions):
            check_results.append(
                'Error: actions must be one of EnterPositionQuantityScaledAction, HedgeAction, ExitPositionAction, '
                'ExitTradeAction, AddTradeAction, AddScaledTradeAction')

        # no duplicate actions
        if not len(set(map(lambda x: type(x), all_actions))) == len(all_actions):
            check_results.append('Error: There are multiple actions of the same type')

        all_child_triggers = reduce(lambda acc, x: acc + x, map(lambda x: x.trigger_requirements.triggers if isinstance(
            x, t.AggregateTriggerRequirements) else [x], strategy.triggers), [])

        for trigger in all_child_triggers:
            if isinstance(trigger, t.PortfolioTrigger):
                continue

            # action one of enter position, exit position, hedge
            if len(trigger.actions) != 1:
                check_results.append('Error: All triggers must contain only 1 action')

            for action in trigger.actions:
                if isinstance(action, (a.EnterPositionQuantityScaledAction, a.AddTradeAction, a.AddScaledTradeAction)):
                    if isinstance(trigger, t.PeriodicTrigger) and \
                            not trigger.trigger_requirements.frequency == action.trade_duration:
                        check_results.append(
                            f'Error: {type(action).__name__}: PeriodicTrigger frequency must be the same '
                            'as trade_duration')
                    if not all((isinstance(p, (EqOption, EqVarianceSwap)))
                               for p in action.priceables):
                        check_results.append(
                            f'Error: {type(action).__name__}: Only EqOption or EqVarianceSwap supported')
                    if isinstance(action, a.EnterPositionQuantityScaledAction):
                        if action.trade_quantity is None or action.trade_quantity_type is None:
                            check_results.append('Error: EnterPositionQuantityScaledAction trade_quantity or '
                                                 'trade_quantity_type is None')
                    if isinstance(action, a.AddScaledTradeAction):
                        if action.scaling_level is None or action.scaling_type is None:
                            check_results.append('Error: AddScaledTradeAction scaling_level or scaling_type is None')
                    expiry_date_modes = map(lambda x: TenorParser(x.expirationDate).get_mode(),
                                            action.priceables)
                    expiry_date_modes = list(set(expiry_date_modes))
                    if len(expiry_date_modes) > 1:
                        check_results.append(
                            f'Error: {type(action).__name__} all priceable expiration_date modifiers must '
                            'be the same. Found [' + ', '.join([str(edm) for edm in expiry_date_modes]) + ']')
                    if expiry_date_modes[0] is not None and expiry_date_modes[0] not in ['otc', 'listed']:
                        check_results.append(
                            f'Error: {type(action).__name__} invalid expiration_date '
                            'modifier ' + expiry_date_modes[0])

                    size_fields = ('quantity', 'number_of_options', 'multiplier')
                    priceable_size_values = [[getattr(p, sf, 1) or 1 for sf in size_fields] for p in action.priceables]
                    priceable_sizes = [reduce(lambda x, y: x * y, size_vals, 1) for size_vals in priceable_size_values]

                    if not all(priceable_size == 1 for priceable_size in priceable_sizes):
                        check_results.append(
                            f'Error: {type(action).__name__} every priceable should have a unit size of 1. '
                            'Found [' + ', '.join([str(s) for s in priceable_sizes]) + ']'
                        )
                elif isinstance(action, a.HedgeAction):
                    if not is_synthetic_forward(action.priceable):
                        check_results.append(
                            'Error: HedgeAction: Hedge instrument must be a synthetic forward - a portfolio of two '
                            'equity options (long call and short put) with the same underlier, strike price and '
                            'expiration date')
                    if not trigger.trigger_requirements.frequency == action.trade_duration:
                        check_results.append(
                            'Error: HedgeAction: PeriodicTrigger frequency must be the same as trade_duration')
                    if not action.risk == EqDelta:
                        check_results.append('Error: HedgeAction: risk type must be EqDelta')
                    if not trigger.trigger_requirements.frequency == '1b':
                        check_results.append('Error: HedgeAction: frequency must be \'1b\'')
                elif isinstance(action, (a.ExitPositionAction, a.ExitTradeAction)):
                    continue
                else:
                    check_results.append('Error: Unsupported action type \'{}\''.format(type(action)))

        return check_results

    @classmethod
    def supports_strategy(cls, strategy):
        check_result = cls.check_strategy(strategy)
        if len(check_result):
            return False

        return True

    @classmethod
    def run_backtest(cls, strategy, start, end, market_model=EquityMarketModel.SFK, cash_accrual=True):
        check_result = cls.check_strategy(strategy)
        if len(check_result):
            raise RuntimeError(check_result)

        underlier_list = None
        roll_frequency = None
        trade_quantity = None
        trade_quantity_type = None
        trade_in_signals = None
        trade_out_signals = None
        hedge = None

        transaction_cost = TransactionCostConfig(None)
        for trigger in strategy.triggers:
            if isinstance(trigger, t.AggregateTrigger):
                child_triggers = trigger.trigger_requirements.triggers

                date_trigger = [x for x in child_triggers if isinstance(x, t.DateTriggerRequirements)][0]
                date_signal = list(map(lambda x: BacktestSignalSeriesItem(x, True),
                                       date_trigger.dates))

                portfolio_trigger = [x for x in child_triggers if isinstance(x, t.PortfolioTriggerRequirements)][0]
                if portfolio_trigger.direction == t.TriggerDirection.EQUAL and \
                        portfolio_trigger.trigger_level == 0:
                    is_trade_in = True
                else:
                    is_trade_in = False

                if is_trade_in:
                    trade_in_signals = date_signal
                else:
                    trade_out_signals = date_signal

            action = trigger.actions[0]
            if isinstance(action, a.EnterPositionQuantityScaledAction):
                underlier_list = cls.__get_underlier_list(action.priceables)
                tp = TenorParser(action.trade_duration)
                roll_frequency = tp.get_date()
                roll_date_mode = tp.get_mode()
                trade_quantity = action.trade_quantity
                trade_quantity_type = action.trade_quantity_type
                expiry_date_mode = TenorParser(action.priceables[0].expiration_date).get_mode()
                transaction_cost.trade_cost_model = TradingCosts(cls.__map_tc_model(action.transaction_cost),
                                                                 cls.__map_tc_model(action.transaction_cost_exit))
            elif isinstance(action, a.AddTradeAction):
                underlier_list = cls.__get_underlier_list(action.priceables)
                tp = TenorParser(action.trade_duration)
                roll_frequency = tp.get_date()
                roll_date_mode = tp.get_mode()
                trade_quantity = 1
                trade_quantity_type = BacktestTradingQuantityType.quantity
                expiry_date_mode = TenorParser(action.priceables[0].expiration_date).get_mode()
                transaction_cost.trade_cost_model = TradingCosts(cls.__map_tc_model(action.transaction_cost),
                                                                 cls.__map_tc_model(action.transaction_cost_exit))
            elif isinstance(action, a.AddScaledTradeAction):
                underlier_list = cls.__get_underlier_list(action.priceables)
                tp = TenorParser(action.trade_duration)
                roll_frequency = tp.get_date()
                roll_date_mode = tp.get_mode()
                trade_quantity = action.scaling_level
                trade_quantity_type = get_backtest_trading_quantity_type(action.scaling_type, action.scaling_risk)
                expiry_date_mode = TenorParser(action.priceables[0].expiration_date).get_mode()
                transaction_cost.trade_cost_model = TradingCosts(cls.__map_tc_model(action.transaction_cost),
                                                                 cls.__map_tc_model(action.transaction_cost_exit))
            elif isinstance(action, a.HedgeAction):
                if trigger.trigger_requirements.frequency == '1b':
                    frequency = 'Daily'
                else:
                    raise RuntimeError('unrecognised hedge frequency')
                hedge = DeltaHedgeParameters(frequency=frequency)
                transaction_cost.hedge_cost_model = TradingCosts(cls.__map_tc_model(action.transaction_cost),
                                                                 cls.__map_tc_model(action.transaction_cost_exit))

        transaction_cost_config = transaction_cost \
            if (transaction_cost.trade_cost_model or transaction_cost.hedge_cost_model) else None

        strategy = StrategySystematic(name="Flow Vol Backtest",
                                      underliers=underlier_list,
                                      index_initial_value=0,
                                      delta_hedge=hedge,
                                      quantity=trade_quantity,
                                      quantity_type=trade_quantity_type,
                                      trade_in_method=TradeInMethod.FixedRoll,
                                      roll_frequency=roll_frequency,
                                      trade_in_signals=trade_in_signals,
                                      trade_out_signals=trade_out_signals,
                                      market_model=market_model,
                                      expiry_date_mode=expiry_date_mode,
                                      roll_date_mode=roll_date_mode,
                                      cash_accrual=cash_accrual,
                                      transaction_cost_config=transaction_cost_config,
                                      use_xasset_backtesting_service=True
                                      )

        result = strategy.backtest(start, end)
        return BacktestResult(result)

    @classmethod
    def __get_underlier_list(cls, priceables):
        priceables_copy = copy.deepcopy(priceables)
        for priceable in priceables_copy:
            edp = TenorParser(priceable.expiration_date)
            priceable.expiration_date = edp.get_date()
            if hasattr(priceable, 'trade_as'):
                expiry_date_mode = get_enum_value(TradeAs, edp.get_mode())
                priceable.trade_as = priceable.trade_as or expiry_date_mode \
                    if isinstance(expiry_date_mode, TradeAs) else None
                print(priceable.trade_as)
        return priceables_copy

    @classmethod
    def __map_tc_model(cls, model: TransactionModel):
        if isinstance(model, ConstantTransactionModel):
            return FixedCostModel(cost=model.cost)
        elif isinstance(model, ScaledTransactionModel):
            if model.scaling_type == EqVega:
                scaling_quantity_type = TransactionCostScalingType.Vega
            else:
                scaling_quantity_type = get_enum_value(TransactionCostScalingType, model.scaling_type)
                if not isinstance(scaling_quantity_type, TransactionCostScalingType):
                    raise RuntimeError(f'unsupported scaled transaction quantity type "{model.scaling_type}"')
            return ScaledCostModel(scaling_quantity_type=scaling_quantity_type, scaling_level=model.scaling_level)
        elif isinstance(model, AggregateTransactionModel):
            return AggregateCostModel(models=[cls.__map_tc_model(m) for m in model.transaction_models],
                                      aggregation_type=CostAggregationType(model.aggregate_type.value))
        return None


class TenorParser(object):

    # match expiration dates expressed as 3m@listed
    expiry_regex = '(.*)@(.*)'

    def __init__(self, expiry: str):
        self.expiry = expiry

    def get_date(self):
        parts = re.search(self.expiry_regex, self.expiry)
        if parts:
            return parts.group(1)
        else:
            return self.expiry

    def get_mode(self):
        parts = re.search(self.expiry_regex, self.expiry)
        if parts:
            return parts.group(2)
        else:
            return None



"""
Copyright 2024 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

from gs_quant.api.gs.scenarios import GsFactorScenarioApi
from gs_quant.markets.factor import Factor
from gs_quant.target.risk import Scenario as TargetScenario, FactorScenarioType
from gs_quant.entities.entitlements import Entitlements
from gs_quant.errors import MqValueError
from enum import Enum
from typing import List, Union, Dict
from pydash import get
from copy import deepcopy

import datetime as dt
import pandas as pd


class ScenarioCalculationType(Enum):
    FACTOR_SCENARIO = "Factor Scenario"


class FactorShock:
    """ Marquee Factor Shock """

    def __init__(self,
                 factor: Union[str, Factor],
                 shock: float):
        self.__factor = factor
        self.__shock = shock

    def __eq__(self, other) -> bool:
        if not isinstance(other, FactorShock):
            return False

        factor_name = self.factor.name if isinstance(self.factor, Factor) else self.factor
        other_factor_name = other.factor.name if isinstance(other.factor, Factor) else other.factor

        return factor_name == other_factor_name and self.shock == other.shock

    def __repr__(self):
        return '%r(factor=%r, shock=%r)' % (self.__class__.__name__,
                                            self.factor,
                                            self.shock)

    @property
    def factor(self) -> Union[str, Factor]:
        """ Get factor being shocked"""
        return self.__factor

    @property
    def shock(self) -> float:
        """ Get factor being shocked"""
        return self.__shock

    @shock.setter
    def shock(self, shock: float):
        self.__shock = shock

    def to_dict(self):
        return {
            "factor": self.factor.name if isinstance(self.factor, Factor) else self.factor,
            "shock": self.shock
        }

    @classmethod
    def from_dict(cls, obj):
        return FactorShock(factor=obj.get("factor"), shock=obj.get("shock"))


class FactorShockParameters:
    def __init__(self,
                 factor_shocks: List[FactorShock] = None,
                 propagate_shocks: bool = None,
                 risk_model: str = None):
        self.__factor_shocks = factor_shocks
        self.__propagate_shocks = propagate_shocks
        self.__risk_model = risk_model

    def __eq__(self, other) -> bool:
        if not isinstance(other, FactorShockParameters):
            return False

        return \
            self.factor_shocks == other.factor_shocks \
            and self.propagate_shocks == other.propagate_shocks and self.risk_model == other.risk_model

    def __repr__(self):
        return '%r(risk_model=%r, propagate_shocks=%r, factor_shocks=%r)' % (self.__class__.__name__,
                                                                             self.risk_model,
                                                                             self.propagate_shocks,
                                                                             self.factor_shocks)

    @property
    def factor_shocks(self) -> List[FactorShock]:
        return self.__factor_shocks

    @factor_shocks.setter
    def factor_shocks(self, factor_shocks: Union[List[FactorShock], Dict, pd.DataFrame]):
        if isinstance(factor_shocks, pd.DataFrame):
            factor_shocks_as_dict = factor_shocks.to_dict(orient='split')
            self.__factor_shocks = [FactorShock(factor=f, shock=s) for f, s in
                                    zip(factor_shocks_as_dict.get('columns'), factor_shocks_as_dict.get('data'))]
        elif isinstance(factor_shocks, Dict):
            self.__factor_shocks = [FactorShock(factor=k, shock=v) for k, v in factor_shocks.items()]
        else:
            self.__factor_shocks = factor_shocks

    @property
    def propagate_shocks(self) -> bool:
        return self.__propagate_shocks

    @propagate_shocks.setter
    def propagate_shocks(self, propagate_shocks: bool):
        self.__propagate_shocks = propagate_shocks

    @property
    def risk_model(self) -> str:
        return self.__risk_model

    @classmethod
    def from_dict(cls, obj: Dict) -> 'FactorShockParameters':
        return cls(factor_shocks=[FactorShock.from_dict(f_shock) for f_shock in obj.get('factorShocks')],
                   risk_model=obj.get("riskModel"),
                   propagate_shocks=obj.get("propagateShocks"))

    def to_dict(self) -> Dict:
        return {
            "riskModel": self.risk_model,
            "propagateShocks": self.propagate_shocks,
            "factorShocks": [f_shock.to_dict() for f_shock in self.factor_shocks]
        }


class HistoricalSimulationParameters:
    def __init__(self,
                 start_date: dt.date = None,
                 end_date: dt.date = None):
        self.__start_date = start_date
        self.__end_date = end_date

    def __eq__(self, other) -> bool:
        if not isinstance(other, HistoricalSimulationParameters):
            return False

        return self.start_date == other.start_date and self.end_date == other.end_date

    def __repr__(self):
        return '%r(start_date=%r, end_date=%r)' % (self.__class__.__name__, self.start_date, self.end_date)

    @property
    def start_date(self) -> dt.date:
        return self.__start_date

    @start_date.setter
    def start_date(self, start_date: dt.date):
        self.__start_date = start_date

    @property
    def end_date(self) -> dt.date:
        return self.__end_date

    @end_date.setter
    def end_date(self, end_date: dt.date):
        self.__end_date = end_date

    @classmethod
    def from_dict(cls, obj: Dict) -> 'HistoricalSimulationParameters':
        return cls(start_date=dt.datetime.strptime(obj.get('startDate'), "%Y-%m-%d").date(),
                   end_date=dt.datetime.strptime(obj.get('endDate'), "%Y-%m-%d").date())

    def to_dict(self) -> Dict:
        return {"startDate": self.start_date, "endDate": self.end_date}


ScenarioParameters = Union[FactorShockParameters, HistoricalSimulationParameters]


class FactorScenario:
    """ Marquee Factor-based Scenario """

    def __init__(self,
                 name: str,
                 type: Union[str, FactorScenarioType],
                 parameters: Union[Dict, HistoricalSimulationParameters, FactorShockParameters],
                 entitlements: Union[Dict, Entitlements] = None,
                 id_: str = None,
                 description: str = None,
                 tags: List[str] = None):
        self.__id = id_
        self.__name = name
        self.__type = type
        self.__description = description
        self.__parameters = parameters \
            if any([isinstance(parameters, FactorShockParameters),
                    isinstance(parameters, HistoricalSimulationParameters)]) \
            else FactorShockParameters.from_dict(parameters) if type == FactorScenarioType.Factor_Shock \
            else HistoricalSimulationParameters.from_dict(parameters)
        self.__entitlements = entitlements
        self.__tags = tags

    def __repr__(self):
        instance_repr = '%r(id=%r, name=%r, description=%r, type=%r, parameters=%r, entitlements=%r, tags=%r)' % (
            self.__class__.__name__,
            self.id,
            self.name,
            self.description,
            self.type,
            self.parameters.__repr__(),
            self.entitlements.__repr__(),
            self.tags)

        return instance_repr

    def __str__(self):
        s = "{}('id={}', 'name={}', 'description={}', 'type={}', 'parameters={}'".format(self.__class__.__name__,
                                                                                         self.id,
                                                                                         self.name,
                                                                                         self.description,
                                                                                         self.type.value,
                                                                                         self.parameters.__repr__())

        s += ")"
        return s

    @property
    def id(self) -> str:
        return self.__id

    @property
    def name(self) -> str:
        return self.__name

    @name.setter
    def name(self, name: str):
        self.__name = name

    @property
    def type(self) -> Union[str, FactorScenarioType]:
        return self.__type

    @property
    def description(self) -> str:
        return self.__description

    @description.setter
    def description(self, description: str):
        self.__description = description

    @property
    def parameters(self) -> ScenarioParameters:
        return self.__parameters

    @parameters.setter
    def parameters(self, parameters: ScenarioParameters):
        self.__parameters = parameters

    @property
    def entitlements(self) -> Entitlements:
        return self.__entitlements

    @entitlements.setter
    def entitlements(self, entitlements: Union[Dict, Entitlements]):
        self.__entitlements = entitlements

    @property
    def tags(self) -> List[str]:
        return self.__tags

    @tags.setter
    def tags(self, tags: List[str]):
        self.__tags = tags

    @classmethod
    def from_dict(cls, scenario_as_dict: Dict) -> 'FactorScenario':
        scenario_data = {
            "name": scenario_as_dict.get('name'),
            "description": scenario_as_dict.get('description'),
            "id_": scenario_as_dict.get('id'),
            "type": scenario_as_dict.get('type'),
            "parameters": get(scenario_as_dict, 'parameters', None),
            "entitlements": get(scenario_as_dict, 'entitlements', None),
            "tags": scenario_as_dict.get('tags')
        }
        return cls(**scenario_data)

    @classmethod
    def from_target(cls, target_scenario: TargetScenario):
        parameters = FactorShockParameters.from_dict(target_scenario.parameters) \
            if target_scenario.type == FactorScenarioType.Factor_Shock else \
            HistoricalSimulationParameters.from_dict(target_scenario.parameters)
        scenario = cls(id_=target_scenario.id,
                       name=target_scenario.name,
                       type=target_scenario.type,
                       parameters=parameters,
                       entitlements=Entitlements.from_target(target_scenario.entitlements),
                       description=target_scenario.description,
                       tags=target_scenario.tags)

        return scenario

    @classmethod
    def get(cls, scenario_id: str) -> 'FactorScenario':
        """
        Get a scenario by its name

        :param scenario_id: The ID of the scenario

        :return: Instance of factor scenario

        :func:`get_by_name` :func:`get_many`
        """
        scenario = GsFactorScenarioApi.get_scenario(scenario_id)
        return cls.from_target(scenario)

    @classmethod
    def get_by_name(cls, scenario_name: str) -> 'FactorScenario':
        """
        Get a scenario by its name

        :param scenario_name: The name of the scenario

        :return: Instance of factor scenario

        :func:`get_many` :func:`get`
        """
        scenario = GsFactorScenarioApi.get_scenario_by_name(scenario_name)
        return cls.from_target(scenario)

    @classmethod
    def get_many(cls,
                 ids: List[str] = None,
                 names: List[str] = None,
                 type: Union[str, FactorScenarioType] = None,
                 risk_model: str = None,
                 shocked_factors: List[str] = None,
                 shocked_factor_categories: List[str] = None,
                 propagated_shocks: bool = None,
                 start_date: dt.date = None,
                 end_date: dt.date = None,
                 tags: List[str] = None,
                 limit: int = 100) -> List['FactorScenario']:
        """
        Get many factor scenarios from Marquee

        :param ids: List of unique Marquee Identifiers
        :param names: Names of requested scenarios
        :param type: The type of the scenario
        :param risk_model: Filter by the risk model
        :param shocked_factors: Return scenarios that shock requested factors.
        :param shocked_factor_categories: Filter scenarios that shock factors in the requested factor categories
        :param propagated_shocks: Return factor shock scenarios that propagate initial factor shocks to other
            non-shocked factors
        :param start_date: Return historical simulation scenarios with an start date that is greater or equal than this
            date. This is only applicable to historical simulation
        :param end_date: Return historical simulation scenarios with an end date that is lower or equal than this date.
            This is only applicable to historical simulation
        :param tags: Returns scenario with any of the requested tags
        :param limit: limit of number of models in response

        :return: list of Factor Shock scenario

        **Usage**

        >>> many_factor_shock_scenarios = FactorScenario.get_many(risk_model="MODEL_ID",
        ...                                                      propagated_shocks=True,
        ...                                                      type="Factor Shock")

        >>> value_shocks_scenarios = FactorScenario.get_many(risk_model="AXIOMA_AXWW4M",
        ...                                                  shocked_factors=["Value"],
        ...                                                  type="Factor Shock")

        >>> many_historical_simulation_scenarios = FactorScenario.get_many(type="Factor Historical Simulation",
        ...                                                                start_date=dt.date(2010, 1, 1),
        ...                                                                end_date=dt.date(2024, 1, 1))

        **See also**

        :func:`get_by_name` :func:`get`
        """
        many_scenarios_as_dict = GsFactorScenarioApi.get_many_scenarios(
            ids=ids,
            names=names,
            type=type.value if isinstance(type, FactorScenarioType) else type,
            risk_model=risk_model,
            shocked_factors=shocked_factors,
            shocked_factor_categories=shocked_factor_categories,
            start_date=start_date,
            end_date=end_date,
            tags=tags,
            limit=limit)

        all_scenarios = [cls.from_target(scenario_as_target) for scenario_as_target in many_scenarios_as_dict]

        if propagated_shocks is not None:
            return [scenario for scenario in all_scenarios if
                    (scenario.parameters.propagate_shocks == propagated_shocks) or
                    scenario.type != FactorScenarioType.Factor_Shock]

        return all_scenarios

    def save(self):
        """ Update factor scenario or Create it if it does not exist"""
        target_scenario = TargetScenario(name=self.name,
                                         type_=self.type,
                                         description=self.description if self.description else None,
                                         parameters=self.parameters.to_dict(),
                                         entitlements=self.entitlements.to_target() if self.entitlements else None,
                                         tags=tuple(self.tags) if self.tags else ())

        if self.id:
            target_scenario.id_ = self.id
            GsFactorScenarioApi.update_scenario(target_scenario)
        else:
            scenario = GsFactorScenarioApi.create_scenario(target_scenario)
            self.__id = scenario.id

    def delete(self):
        """Deletes factor scenario from Marquee"""
        if not self.id:
            raise MqValueError("Cannot delete scenario that has not been created in Marquee")

        GsFactorScenarioApi.delete_scenario(self.id)

    def clone(self):
        """
        Clones an existing scenario

        :return: New scenario instance with identical parameters

        **Usage**


        >>> from gs_quant.markets.scenario import FactorScenario
        >>>
        >>> initial_scenario = FactorScenario.get("SCENARIO_ID")
        >>> clone = initial_scenario.clone()
        >>> clone.save()

        **See also**

        :func:`save`
        """
        parameters = deepcopy(self.parameters)

        return FactorScenario(name=f"{self.name} copy", description=self.description,
                              type=self.type, parameters=parameters)


Scenario = Union[FactorScenario]



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

from gs_quant.base import *
from gs_quant.common import *
import datetime
from typing import Dict, Optional, Tuple, Union
from dataclasses import dataclass, field
from dataclasses_json import LetterCase, config, dataclass_json
from enum import Enum


class Encoding(EnumBase, Enum):    
    
    HTML = 'HTML'
    URL = 'URL'
    Unicode = 'Unicode'
    Base64 = 'Base64'
    Hex = 'Hex'
    ASCII = 'ASCII'    


class HedgeModel(EnumBase, Enum):    
    
    Smile = 'Smile'
    BlackScholes = 'BlackScholes'    


class ImgType(EnumBase, Enum):    
    
    APNG = 'APNG'
    AVIF = 'AVIF'
    GIF = 'GIF'
    JPEG = 'JPEG'
    PNG = 'PNG'
    SVG = 'SVG'
    WEBP = 'WEBP'    


class MarketRefOverrideType(EnumBase, Enum):    
    
    """Market Ref Override Type Spot or Fwd"""

    Spot = 'Spot'
    Fwd = 'Fwd'    


class OverlayType(EnumBase, Enum):    
    
    """Type"""

    Payout = 'Payout'
    Price = 'Price'
    MtM = 'MtM'
    Delta = 'Delta'
    Vega = 'Vega'
    Theta = 'Theta'
    RelativeCheapness = 'RelativeCheapness'
    ProbabilityDistribution = 'ProbabilityDistribution'
    RealisedProbability = 'RealisedProbability'
    MacroEvents = 'MacroEvents'
    MicroEvents = 'MicroEvents'
    Gamma = 'Gamma'
    _None = 'None'    


class PriceFormat(EnumBase, Enum):    
    
    """display unit for price"""

    Absolute = 'Absolute'
    Relative = 'Relative'
    Cents = 'Cents'    


class RelativeExpiryType(EnumBase, Enum):    
    
    Relative = 'Relative'
    Fixed = 'Fixed'    


class RelativeStrikeType(EnumBase, Enum):    
    
    Relative_Delta = 'Relative Delta'
    Relative_Spot = 'Relative Spot'
    Relative_Fwd = 'Relative Fwd'
    Fixed = 'Fixed'    


@dataclass
class HedgeTypes(Base):
    pass


@handle_camel_case_args
@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(unsafe_hash=True, repr=False)
class CustomDeltaHedge(HedgeTypes):
    amount: float = field(default=None, metadata=field_metadata)
    type_: Optional[str] = field(init=False, default='CustomDeltaHedge', metadata=config(field_name='type', exclude=exclude_none))
    name: Optional[str] = field(default=None, metadata=name_metadata)


class GenericResponse(DictBase):
    pass


@handle_camel_case_args
@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(unsafe_hash=True, repr=False)
class HyperLinkImageComments(CustomComments):
    url: Optional[str] = field(default=None, metadata=field_metadata)
    comment_type: Optional[str] = field(init=False, default='hyperLinkImageComments', metadata=field_metadata)
    name: Optional[str] = field(default=None, metadata=name_metadata)


@handle_camel_case_args
@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(unsafe_hash=True, repr=False)
class SalesPremiumAdjustment(Base):
    value: Optional[float] = field(default=None, metadata=field_metadata)
    unit: Optional[str] = field(default=None, metadata=field_metadata)
    name: Optional[str] = field(default=None, metadata=name_metadata)


@handle_camel_case_args
@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(unsafe_hash=True, repr=False)
class SolvingTarget(Base):
    constraint: Optional[float] = field(default=None, metadata=field_metadata)
    name: Optional[str] = field(default=None, metadata=name_metadata)


@handle_camel_case_args
@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(unsafe_hash=True, repr=False)
class WorkflowEntitlement(Base):
    type_: str = field(default=None, metadata=config(field_name='type', exclude=exclude_none))
    value: str = field(default=None, metadata=field_metadata)
    name: Optional[str] = field(default=None, metadata=name_metadata)


@handle_camel_case_args
@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(unsafe_hash=True, repr=False)
class BinaryImageComments(CustomComments):
    data: Optional[str] = field(default=None, metadata=field_metadata)
    img_type: Optional[ImgType] = field(default=None, metadata=field_metadata)
    encoding: Optional[Encoding] = field(default=None, metadata=field_metadata)
    comment_type: Optional[str] = field(init=False, default='binaryImageComments', metadata=field_metadata)
    name: Optional[str] = field(default=None, metadata=name_metadata)


@handle_camel_case_args
@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(unsafe_hash=True, repr=False)
class ChartingParameters(Base):
    spot_style: Optional[str] = field(default=None, metadata=field_metadata)
    overlay: Optional[OverlayType] = field(default=None, metadata=field_metadata)
    underlay: Optional[OverlayType] = field(default=None, metadata=field_metadata)
    description: Optional[str] = field(default=None, metadata=field_metadata)
    name: Optional[str] = field(default=None, metadata=name_metadata)


@handle_camel_case_args
@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(unsafe_hash=True, repr=False)
class DeltaHedge(HedgeTypes):
    model: Optional[HedgeModel] = field(default=None, metadata=field_metadata)
    type_: Optional[str] = field(init=False, default='DeltaHedge', metadata=config(field_name='type', exclude=exclude_none))
    name: Optional[str] = field(default=None, metadata=name_metadata)


@handle_camel_case_args
@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(unsafe_hash=True, repr=False)
class MarketRefOverride(Base):
    asset: Optional[str] = field(default=None, metadata=field_metadata)
    value: Optional[float] = field(default=None, metadata=field_metadata)
    mkt_ref_override_type: Optional[MarketRefOverrideType] = field(default=None, metadata=field_metadata)
    ref_date: Optional[datetime.date] = field(default=None, metadata=field_metadata)
    name: Optional[str] = field(default=None, metadata=name_metadata)


@handle_camel_case_args
@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(unsafe_hash=True, repr=False)
class OverlayParameters(Base):
    overlay_type: Optional[OverlayType] = field(default=None, metadata=field_metadata)
    name: Optional[str] = field(default=None, metadata=name_metadata)


@handle_camel_case_args
@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(unsafe_hash=True, repr=False)
class SolvingInfo(Base):
    target: Optional[SolvingTarget] = field(default=None, metadata=field_metadata)
    name: Optional[str] = field(default=None, metadata=name_metadata)


@handle_camel_case_args
@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(unsafe_hash=True, repr=False)
class StrategyDescription(Base):
    strategy_type: Optional[str] = field(default=None, metadata=field_metadata)
    long_short: Optional[LongShort] = field(default=LongShort.Long, metadata=field_metadata)
    assets: Optional[Tuple[str, ...]] = field(default=None, metadata=field_metadata)
    asset_classes: Optional[Tuple[str, ...]] = field(default=None, metadata=field_metadata)
    name: Optional[str] = field(default=None, metadata=name_metadata)


@handle_camel_case_args
@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(unsafe_hash=True, repr=False)
class WorkflowEntitlements(Base):
    version: Optional[int] = field(default=None, metadata=field_metadata)
    readers: Optional[Tuple[WorkflowEntitlement, ...]] = field(default=None, metadata=field_metadata)
    writers: Optional[Tuple[WorkflowEntitlement, ...]] = field(default=None, metadata=field_metadata)
    name: Optional[str] = field(default=None, metadata=name_metadata)


@handle_camel_case_args
@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(unsafe_hash=True, repr=False)
class MarketDataParameters(Base):
    max_history: Optional[datetime.date] = field(default=None, metadata=field_metadata)
    timestamp: Optional[datetime.datetime] = field(default=None, metadata=field_metadata)
    spot_ref: Optional[float] = field(default=None, metadata=field_metadata)
    mkt_ref_override: Optional[Tuple[MarketRefOverride, ...]] = field(default=None, metadata=field_metadata)
    name: Optional[str] = field(default=None, metadata=name_metadata)


@handle_camel_case_args
@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(unsafe_hash=True, repr=False)
class VisualStructuringReport(QuoteReport):
    report_type: Optional[str] = field(default='VisualStructuringReport', metadata=field_metadata)
    position_set_id: Optional[str] = field(default=None, metadata=field_metadata)
    quick_entry_text: Optional[str] = field(default=None, metadata=field_metadata)
    as_of_date: Optional[datetime.date] = field(default=None, metadata=field_metadata)
    market_data_parameters: Optional[MarketDataParameters] = field(default=None, metadata=field_metadata)
    overlay_parameters: Optional[OverlayParameters] = field(default=None, metadata=field_metadata)
    solving_info: Optional[SolvingInfo] = field(default=None, metadata=field_metadata)
    charting_parameters: Optional[ChartingParameters] = field(default=None, metadata=field_metadata)
    comments: Optional[Tuple[CustomComments, ...]] = field(default=None, metadata=field_metadata)
    strategy_description: Optional[StrategyDescription] = field(default=None, metadata=field_metadata)
    asset_class: Optional[str] = field(default=None, metadata=field_metadata)
    hedge_instruction: Optional[HedgeTypes] = field(default=None, metadata=field_metadata)
    sales_premium_adjustment: Optional[SalesPremiumAdjustment] = field(default=None, metadata=field_metadata)
    price_format: Optional[PriceFormat] = field(default=None, metadata=field_metadata)
    strike_and_barrier_type: Optional[RelativeStrikeType] = field(default=None, metadata=field_metadata)
    expiry_type: Optional[RelativeExpiryType] = field(default=None, metadata=field_metadata)
    name: Optional[str] = field(default=None, metadata=name_metadata)


@handle_camel_case_args
@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(unsafe_hash=True, repr=False)
class SaveQuoteRequest(Base):
    positions: Tuple[PositionSet, ...] = field(default=None, metadata=field_metadata)
    measures: Tuple[RiskMeasure, ...] = field(default=None, metadata=field_metadata)
    pricing_and_market_data_as_of: Optional[Tuple[PricingDateAndMarketDataAsOf, ...]] = field(default=None, metadata=field_metadata)
    pricing_location: Optional[PricingLocation] = field(default=None, metadata=field_metadata)
    scenario: Optional[MarketDataScenario] = field(default=None, metadata=field_metadata)
    parameters: Optional[RiskRequestParameters] = field(default=None, metadata=field_metadata)
    reports: Optional[Tuple[QuoteReport, ...]] = field(default=None, metadata=field_metadata)
    shared_users: Optional[Tuple[str, ...]] = field(default=None, metadata=field_metadata)
    comments: Optional[str] = field(default=None, metadata=field_metadata)
    description: Optional[str] = field(default=None, metadata=field_metadata)
    original_workflow_id: Optional[str] = field(default=None, metadata=field_metadata)
    is_sharing_parent: Optional[bool] = field(default=None, metadata=field_metadata)
    entitlements: Optional[WorkflowEntitlements] = field(default=None, metadata=field_metadata)
    name: Optional[str] = field(default=None, metadata=name_metadata)


@handle_camel_case_args
@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(unsafe_hash=True, repr=False)
class WorkflowPosition(Base):
    id_: str = field(default=None, metadata=config(field_name='id', exclude=exclude_none))
    position_sets: Optional[Tuple[PositionSet, ...]] = field(default=None, metadata=field_metadata)
    reports: Optional[Tuple[QuoteReport, ...]] = field(default=None, metadata=field_metadata)
    comments: Optional[str] = field(default=None, metadata=field_metadata)
    original_workflow_id: Optional[str] = field(default=None, metadata=field_metadata)
    description: Optional[str] = field(default=None, metadata=field_metadata)
    entitlements: Optional[WorkflowEntitlements] = field(default=None, metadata=field_metadata)
    creator: Optional[str] = field(default=None, metadata=field_metadata)
    originating_system: Optional[str] = field(default=None, metadata=field_metadata)
    is_read_only: Optional[bool] = field(default=None, metadata=field_metadata)
    name: Optional[str] = field(default=None, metadata=name_metadata)


@handle_camel_case_args
@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(unsafe_hash=True, repr=False)
class WorkflowPositionsResponse(Base):
    results: Optional[Tuple[WorkflowPosition, ...]] = field(default=None, metadata=field_metadata)
    name: Optional[str] = field(default=None, metadata=name_metadata)



"""
Copyright 2020 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import datetime as dt
import re
from enum import Enum
from typing import Optional, Union, List

import pandas as pd
import numpy as np
import math
import scipy.stats as stats
from pandas.tseries.offsets import BDay
from pydash import decapitalize

from gs_quant.api.gs.data import QueryType, GsDataApi, DataQuery
from gs_quant.data import DataMeasure
from gs_quant.data.core import DataContext
from gs_quant.entities.entity import EntityType
from gs_quant.errors import MqValueError
from gs_quant.markets.portfolio_manager import PortfolioManager
from gs_quant.markets.report import FactorRiskReport, PerformanceReport, ThematicReport, ReturnFormat, \
    format_aum_for_return_calculation, get_pnl_percent, get_factor_pnl_percent_for_single_factor
from gs_quant.markets.securities import Bond
from gs_quant.models.risk_model import FactorRiskModel
from gs_quant.target.reports import PositionSourceType
from gs_quant.timeseries import plot_measure_entity, beta, correlation, max_drawdown
from gs_quant.timeseries.algebra import geometrically_aggregate
from gs_quant.timeseries.measures import _extract_series_from_df, SecurityMaster, AssetIdentifier


class Unit(Enum):
    NOTIONAL = 'Notional'
    PERCENT = 'Percent'


@plot_measure_entity(EntityType.REPORT, [QueryType.FACTOR_EXPOSURE])
def factor_exposure(report_id: str, factor_name: str, unit: str = 'Notional', *, source: str = None,
                    real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Factor exposure data associated with a factor in a factor risk report

    :param report_id: factor risk report id
    :param factor_name: factor name
    :param unit: unit in which the timeseries is returned (Notional or Percent)
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: Timeseries of factor exposure for requested factor
    """
    return _get_factor_data(report_id, factor_name, QueryType.FACTOR_EXPOSURE, Unit(unit))


@plot_measure_entity(EntityType.REPORT, [QueryType.FACTOR_PNL])
def factor_pnl(report_id: str, factor_name: str, unit: str = 'Notional', *, source: str = None,
               real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Factor PnL data associated with a factor in a factor risk report

    :param report_id: factor risk report id
    :param factor_name: factor name
    :param unit: unit in which the timeseries is returned (Notional or Percent)
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: Timeseries of factor pnl for requested factor
    """
    return _get_factor_data(report_id, factor_name, QueryType.FACTOR_PNL, Unit(unit))


@plot_measure_entity(EntityType.REPORT, [QueryType.FACTOR_PROPORTION_OF_RISK])
def factor_proportion_of_risk(report_id: str, factor_name: str, *, source: str = None,
                              real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Factor proportion of risk data associated with a factor in a factor risk report

    :param report_id: factor risk report id
    :param factor_name: factor name
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: Timeseries of factor proportion of risk for requested factor
    """
    return _get_factor_data(report_id, factor_name, QueryType.FACTOR_PROPORTION_OF_RISK)


@plot_measure_entity(EntityType.REPORT, [QueryType.DAILY_RISK])
def daily_risk(report_id: str, factor_name: str = 'Total', *, source: str = None,
               real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Daily risk data associated with a factor in a factor risk report

    :param report_id: factor risk report id
    :param factor_name: factor name (must be "Factor", "Specific", or "Total")
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: Timeseries of daily risk for requested factor
    """
    return _get_factor_data(report_id, factor_name, QueryType.DAILY_RISK)


@plot_measure_entity(EntityType.REPORT, [QueryType.ANNUAL_RISK])
def annual_risk(report_id: str, factor_name: str = 'Total', *, source: str = None,
                real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Annual risk data associated with a factor in a factor risk report

    :param report_id: factor risk report id
    :param factor_name: factor name (must be "Factor", "Specific", or "Total")
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: Timeseries of daily risk for requested factor
    """
    return _get_factor_data(report_id, factor_name, QueryType.ANNUAL_RISK)


@plot_measure_entity(EntityType.REPORT, [QueryType.PNL])
def normalized_performance(report_id: str, leg: str = None, *, source: str = None,
                           real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Returns the Normalized Performance of a performance report based on AUM source
    :param report_id: id of performance report
    :param leg: short or long
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: portfolio normalized performance

    **Usage**

    Returns the normalized performance of the portfolio.

    :math:`NP(L/S)_{t} = SUM( PNL(L/S)_{t}/ ( EXP(L/S)_{t} ) - cPNL(L/S)_{t-1) )
        if ( EXP(L/S)_{t} ) > 0
        else:
            1/ SUM( PNL(L/S)_{t}/ ( EXP(L/S)_{t} ) - cPNL(L/S)_{t-1) )`
    For each leg, short and long, then:
    :math:`NP_{t} = NP(L)_{t} * SUM(EXP(L)) / SUM(GROSS_EXP) + NP(S)_{t} * SUM(EXP(S)) / SUM(GROSS_EXP) + 1`

    If leg is short, set SUM(EXP(L)) to 0, if leg is long, set SUM(EXP(S)) to 0

    where :math:`cPNL(L/S)_{t-1}` is your performance reports cumulative long or short PNL at date t-1
    where :math:`PNL(L/S)_{t}` is your performance reports long or short pnl at date t
    where :math:`GROSS_EXP_{t}` is portfolio gross exposure on date t
    where :math:`EXP(L/S)_{t}` is the long or short exposure on date t

    """
    start_date = DataContext.current.start_time
    end_date = DataContext.current.end_time

    start_date = (start_date - BDay(1)).date()
    end_date = end_date.date()

    performance_report = PerformanceReport.get(report_id)

    constituent_data = performance_report.get_portfolio_constituents(
        fields=['assetId', 'pnl', 'quantity', 'netExposure'], start_date=start_date, end_date=end_date).set_index(
        'date')

    if leg:
        if leg.lower() == "long":
            constituent_data = constituent_data[constituent_data['quantity'] > 0]
        if leg.lower() == "short":
            constituent_data = constituent_data[constituent_data['quantity'] < 0]

    # Split into long and short and aggregate across dates
    long_side = _return_metrics(constituent_data[constituent_data['quantity'] > 0],
                                list(constituent_data.index.unique()), "long")
    short_side = _return_metrics(constituent_data[constituent_data['quantity'] < 0],
                                 list(constituent_data.index.unique()), "short")

    short_exposure = sum(abs(short_side['exposure']))
    long_exposure = sum(long_side['exposure'])
    gross_exposure = short_exposure + long_exposure

    long_side['longRetWeighted'] = (long_side['longMetrics'] - 1) * (long_exposure / gross_exposure)
    short_side['shortRetWeighted'] = (short_side['shortMetrics'] - 1) * (short_exposure / gross_exposure)

    combined = long_side[['longRetWeighted']].join(short_side[['shortRetWeighted']], how='inner')
    combined['normalizedPerformance'] = combined['longRetWeighted'] + combined['shortRetWeighted'] + 1
    return pd.Series(combined['normalizedPerformance'], name="normalizedPerformance").dropna()


@plot_measure_entity(EntityType.REPORT, [QueryType.PNL])
def long_pnl(report_id: str, *, source: str = None,
             real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    PNL from long holdings

    :param report_id: id of performance report
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: portfolio long pnl
    """
    start_date = DataContext.current.start_time.date()
    end_date = DataContext.current.end_time.date()
    performance_report = PerformanceReport.get(report_id)

    constituent_data = performance_report.get_portfolio_constituents(
        fields=['pnl', 'quantity'], start_date=start_date, end_date=end_date).set_index('date')
    long_leg = constituent_data[constituent_data['quantity'] > 0]['pnl']
    long_leg = long_leg.groupby(level=0).sum()
    return pd.Series(long_leg, name="longPnl")


@plot_measure_entity(EntityType.REPORT, [QueryType.PNL])
def short_pnl(report_id: str, *, source: str = None,
              real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """

    PNL from short holdings
    :param report_id: id of performance report
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: portfolio short pnl
    """
    start_date = DataContext.current.start_time.date()
    end_date = DataContext.current.end_time.date()
    performance_report = PerformanceReport.get(report_id)

    constituent_data = performance_report.get_portfolio_constituents(
        fields=['pnl', 'quantity'], start_date=start_date, end_date=end_date).set_index('date')
    short_leg = constituent_data[constituent_data['quantity'] < 0]['pnl']
    short_leg = short_leg.groupby(level=0).sum()
    return pd.Series(short_leg, name="shortPnl")


@plot_measure_entity(EntityType.REPORT, [QueryType.THEMATIC_EXPOSURE])
def thematic_exposure(report_id: str, basket_ticker: str, *, source: str = None,
                      real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Thematic exposure of a portfolio to a requested GS thematic flagship basket

    :param report_id: portfolio thematic analytics report id
    :param basket_ticker: ticker for thematic basket
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: Timeseries of daily thematic beta of portfolio to requested flagship basket
    """
    thematic_report = ThematicReport.get(report_id)
    asset = SecurityMaster.get_asset(basket_ticker, AssetIdentifier.TICKER)
    df = thematic_report.get_thematic_exposure(start_date=DataContext.current.start_date,
                                               end_date=DataContext.current.end_date,
                                               basket_ids=[asset.get_marquee_id()])
    if not df.empty:
        df = df.set_index('date')
        df.index = pd.to_datetime(df.index)
    return _extract_series_from_df(df, QueryType.THEMATIC_EXPOSURE)


@plot_measure_entity(EntityType.REPORT, [QueryType.THEMATIC_EXPOSURE])
def thematic_beta(report_id: str, basket_ticker: str, *, source: str = None,
                  real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Thematic beta values of a portfolio to a requested GS thematic flagship basket

    :param report_id: portfolio thematic analytics report id
    :param basket_ticker: ticker for thematic basket
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: Timeseries of daily thematic beta of portfolio to requested flagship basket
    """
    thematic_report = ThematicReport.get(report_id)
    asset = SecurityMaster.get_asset(basket_ticker, AssetIdentifier.TICKER)
    df = thematic_report.get_thematic_betas(start_date=DataContext.current.start_date,
                                            end_date=DataContext.current.end_date,
                                            basket_ids=[asset.get_marquee_id()])
    if not df.empty:
        df = df.set_index('date')
        df.index = pd.to_datetime(df.index)
    return _extract_series_from_df(df, QueryType.THEMATIC_BETA)


@plot_measure_entity(EntityType.REPORT, [QueryType.AUM])
def aum(report_id: str, *, source: str = None,
        real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    AUM of the portfolio

    :param report_id: id of performance report
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: portfolio long pnl
    """
    start_date = DataContext.current.start_time.date()
    end_date = DataContext.current.end_time.date()
    performance_report = PerformanceReport.get(report_id)
    aum_curve = performance_report.get_aum(start_date=start_date, end_date=end_date)
    aum_dict = [{'date': key, 'aum': aum_curve[key]} for key in aum_curve]

    # Create and return timeseries
    df = pd.DataFrame(aum_dict)
    if not df.empty:
        df = df.set_index('date')
        df.index = pd.to_datetime(df.index)
    return _extract_series_from_df(df, QueryType.AUM)


@plot_measure_entity(EntityType.REPORT, [QueryType.PNL])
def pnl(report_id: str, unit: str = 'Notional', *, source: str = None,
        real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    PNL from all holdings, if unit is Percent, geometrically aggregate over time frame

    :param report_id: id of performance report
    :param unit: by default uses mode as Notional, but can also be set to Percent
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: portfolio pnl
    """
    start_date = DataContext.current.start_time.date()
    end_date = DataContext.current.end_time.date()
    performance_report = PerformanceReport.get(report_id)
    pnl_df = performance_report.get_pnl(start_date, end_date)

    if unit == Unit.PERCENT.value:
        return get_pnl_percent(performance_report, pnl_df, 'pnl', start_date, end_date)
    else:
        pnl_df = pnl_df.set_index('date')
        return pd.Series(pnl_df['pnl'], name="pnl")


@plot_measure_entity(EntityType.REPORT)
def historical_simulation_estimated_pnl(report_id: str, *, source: str = None,
                                        real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Estimated Pnl from replaying a historical simulation scenario on your latest positions
    :param report_id: id of performance report
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: portfolio estimated pnl
    """

    factor_attributed_pnl = _replay_historical_factor_moves_on_latest_positions(report_id, [])
    total_factor_attributed_pnl = factor_attributed_pnl.apply(np.sum, axis=1).to_frame("estimatedPnl")

    total_factor_attributed_pnl.index = pd.to_datetime(total_factor_attributed_pnl.index)
    return total_factor_attributed_pnl.squeeze()


@plot_measure_entity(EntityType.REPORT)
def historical_simulation_estimated_factor_attribution(report_id: str, factor_name: str, *, source: str = None,
                                                       real_time: bool = False,
                                                       request_id: Optional[str] = None) -> pd.Series:
    """
    Estimated Pnl attributed to the factor after replaying a historical simulation scenario on a portfolio's latest
    positions
    :param report_id: id of performance report
    :param factor_name: name of the factor
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: portfolio estimated pnl
    """

    factor_attributed_pnl = _replay_historical_factor_moves_on_latest_positions(report_id, [factor_name])
    factor_attributed_pnl.index = pd.to_datetime(factor_attributed_pnl.index)

    return factor_attributed_pnl.squeeze()


@plot_measure_entity(EntityType.REPORT)
def hit_rate(report_id: str, rolling_window: Union[int, str], *, source: str = None,
             real_time: bool = False,
             request_id: Optional[str] = None) -> pd.Series:
    """
    The hit rate of a portfolio is a percentage of positions that have generated positive returns over a given period.

    :param report_id: id of performance report
    :param rolling_window: size of rolling window
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of the hit rate
    """
    start_date = DataContext.current.start_time.date()
    end_date = DataContext.current.end_time.date()
    performance_report = PerformanceReport.get(report_id)
    window = _parse_window(rolling_window)
    portfolio_constituents = performance_report.get_portfolio_constituents(fields=['date', 'pnl'],
                                                                           start_date=start_date,
                                                                           end_date=end_date).set_index('date')

    portfolio_constituents = portfolio_constituents[portfolio_constituents['entryType'] == 'Holding']
    portfolio_constituents = portfolio_constituents.sort_values('date')
    pivot_constituents = portfolio_constituents.pivot_table(index='date', columns='assetId', values='pnl')

    rolling_compound = (pivot_constituents).rolling(window=window).sum()
    positive_count = (rolling_compound > 0).sum(axis=1)
    valid_assets = rolling_compound.count(axis=1)

    return positive_count / valid_assets


@plot_measure_entity(EntityType.REPORT)
def portfolio_max_drawdown(report_id: str, rolling_window: Union[int, str], *,
                           source: str = None,
                           real_time: bool = False,
                           request_id: Optional[str] = None) -> pd.Series:
    """
    The largest drop from peak to trough in a sub-period over the stated timeframe

    :param report_id: id of performance report
    :param rolling_window: size of rolling window
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: portfolio estimated pnl
    """
    aum_series = aum(report_id)
    rolling_window = _parse_window(rolling_window)
    max_drawdown_series = aum_series.rolling(window=rolling_window).apply(_max_drawdown, raw=False)

    return max_drawdown_series


@plot_measure_entity(EntityType.REPORT)
def drawdown_length(report_id: str, rolling_window: Union[int, str], *,
                    source: str = None,
                    real_time: bool = False,
                    request_id: Optional[str] = None) -> pd.Series:
    """
    The length in days between the peak and the trough of the maximum drawdown

    :param report_id: id of performance report
    :param rolling_window: size of rolling window
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of the drawdown length
    """
    aum_series = aum(report_id)
    rolling_window = _parse_window(rolling_window)
    drawdown_length_series = aum_series.rolling(window=rolling_window).apply(_drawdown_length, raw=False)

    return drawdown_length_series


@plot_measure_entity(EntityType.REPORT)
def max_recovery_period(report_id: str, rolling_window: Union[int, str], *,
                        source: str = None,
                        real_time: bool = False,
                        request_id: Optional[str] = None) -> pd.Series:
    """
    The maximum number of days used to reach a previously broken price level over the stated timeframe.

    :param report_id: id of performance report
    :param rolling_window: size of rolling window
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of max recovery period
    """
    aum_series = aum(report_id)
    rolling_window = _parse_window(rolling_window)
    recovery_period_series = aum_series.rolling(window=rolling_window).apply(_max_recovery_period, raw=False)

    return recovery_period_series


def _max_recovery_period(series):
    peak = series[0]
    recovery_periods = []
    current_drawdown = 0
    in_drawdown = False

    for val in series[1:]:
        if val < peak:
            in_drawdown = True
            current_drawdown += 1
        else:
            if in_drawdown:
                recovery_periods.append(current_drawdown)
                in_drawdown = False
            peak = val
            current_drawdown = 0
    return max(recovery_periods) if recovery_periods else 0


def _drawdown_length(series):
    peak = series[0]
    drawdown_start = None
    max_drawdown_length = 0
    current_length = 0

    for value in series:
        if value >= peak:
            # Reset when a new peak is found
            peak = value
            drawdown_start = None
            current_length = 0
        else:
            # Track the drawdown length
            if drawdown_start is None:
                drawdown_start = value
            current_length += 1
            max_drawdown_length = max(max_drawdown_length, current_length)

    return max_drawdown_length


@plot_measure_entity(EntityType.REPORT)
def standard_deviation(report_id: str, rolling_window: Union[int, str], *, source: str = None,
                       real_time: bool = False,
                       request_id: Optional[str] = None) -> pd.Series:
    """
    Volatility of the total return over the stated time frame.

    :param report_id: id of performance report
    :param rolling_window: length of window
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of standard deviation
    """
    portfolio_pnl = _get_daily_pnl(report_id)

    rolling_window = _parse_window(rolling_window)
    rolling_std = portfolio_pnl.rolling(window=rolling_window).std()

    return rolling_std


@plot_measure_entity(EntityType.REPORT)
def downside_risk(report_id: str, rolling_window: Union[int, str], *, source: str = None,
                  real_time: bool = False,
                  request_id: Optional[str] = None) -> pd.Series:
    """
    Volatility of the periodic returns that are lower than the mean return over the stated time frame.
    Standard deviation is calculated using all returns, downside risk is calculated using only the returns
    that are below the mean.

    :param report_id: id of performance report
    :param rolling_window: length of window
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of downside_risk
    """
    portfolio_pnl = _get_daily_pnl(report_id)
    rolling_window = _parse_window(rolling_window)
    downside_risk_series = portfolio_pnl.rolling(window=rolling_window).apply(_rolling_downside_risk, raw=False)

    return downside_risk_series


@plot_measure_entity(EntityType.REPORT)
def semi_variance(report_id: str, rolling_window: Union[int, str], *, source: str = None,
                  real_time: bool = False,
                  request_id: Optional[str] = None) -> pd.Series:
    """
    Volatility of the periodic returns that are lower than the mean return over the stated time frame.
    Standard deviation is calculated using all returns, semi variance is calculated using only the returns
    that are below 0.

    :param report_id: id of performance report
    :param rolling_window: length of window
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of downside_risk
    """
    portfolio_pnl = _get_daily_pnl(report_id)
    rolling_window = _parse_window(rolling_window)
    semi_variance_series = portfolio_pnl.rolling(window=rolling_window).apply(_rolling_semi_variance, raw=False)

    return semi_variance_series


@plot_measure_entity(EntityType.REPORT)
def kurtosis(report_id: str, rolling_window: Union[int, str], *, source: str = None,
             real_time: bool = False,
             request_id: Optional[str] = None) -> pd.Series:
    """
    Kurtosis measure the peakedness or flatness of the return distribution over the state time frame.
    In a flat distribution the average value is more likely to occur.


    :param report_id: id of performance report
    :param rolling_window: length of window
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of kurtosis
    """
    portfolio_pnl = _get_daily_pnl(report_id)
    rolling_window = _parse_window(rolling_window)
    rolling_kurtosis = portfolio_pnl.rolling(window=rolling_window).apply(
        lambda x: stats.kurtosis(x, fisher=True, bias=False),
        raw=False
    )

    return rolling_kurtosis


@plot_measure_entity(EntityType.REPORT)
def skewness(report_id: str, rolling_window: Union[int, str], *, source: str = None,
             real_time: bool = False,
             request_id: Optional[str] = None) -> pd.Series:
    """
    Skewness measures the degree of symmetry of the return distribution over the stated time frame.
    If the left tail is more pronounced than the right tail, it is said the return have negative skewness.

    :param report_id: id of performance report
    :param rolling_window: length of window
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of skewness
    """
    portfolio_pnl = _get_daily_pnl(report_id)
    rolling_window = _parse_window(rolling_window)
    rolling_skewness = portfolio_pnl.rolling(window=rolling_window).apply(
        lambda x: stats.skew(x, bias=False),
        raw=False
    )

    return rolling_skewness


@plot_measure_entity(EntityType.REPORT)
def realized_var(report_id: str, rolling_window: Union[int, str],
                 confidence_interval: float = .95, *, source: str = None,
                 real_time: bool = False,
                 request_id: Optional[str] = None) -> pd.Series:
    """
    The maximum expected loss of the portfolios, calculated using
    the natural distribution of returns over a stated time frame and is based on a confidence level.

    :param report_id: id of performance report
    :param rolling_window: length of window
    :param confidence_interval: confidence interval for variance
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of realized var
    """
    portfolio_pnl = _get_daily_pnl(report_id)
    rolling_window = _parse_window(rolling_window)
    realized_variance = portfolio_pnl.rolling(window=rolling_window).quantile(1 - confidence_interval)
    realized_variance = realized_variance * -1

    return realized_variance


@plot_measure_entity(EntityType.REPORT)
def tracking_error(report_id: str, benchmark_id: str, rolling_window: Union[int, str],
                   *, source: str = None,
                   real_time: bool = False,
                   request_id: Optional[str] = None) -> pd.Series:
    """
    The standard deviation of the excess return relative to the benchmark over the state time frame.

    Tracking Error = sqrt( sum((ExcessRi - ExcessRmean)^2) / (N-1) )

    Where:

    ExcessRi = excess return in period i

    ExcessRmean = mean excess return during bull markets

    N = number of periods

    :param report_id: id of performance report
    :param rolling_window: length of window
    :param benchmark_id: id of benchmark
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of tracking error
    """
    start_date = DataContext.current.start_time.date()
    end_date = DataContext.current.end_time.date()
    rolling_window = _parse_window(rolling_window)
    benchmark_returns = _get_benchmark_daily_return(benchmark_id, start_date, end_date)
    portfolio_returns = _get_daily_pnl(report_id)
    active_return = (portfolio_returns - benchmark_returns)
    rolling_error = active_return.rolling(window=rolling_window).std()
    return rolling_error


@plot_measure_entity(EntityType.REPORT)
def tracking_error_bear(report_id: str, benchmark_id: str, rolling_window: Union[int, str], *,
                        source: str = None,
                        real_time: bool = False,
                        request_id: Optional[str] = None) -> pd.Series:
    """
    The standard deviation of the excess return relative to the benchmark over the state time frame,
    only counting the periods when benchmark returns were negative.

    Bear Tracking Error = sqrt( sum((ExcessRi - ExcessRmeanBear)^2) / (N-1) ) for all when Benchmark Return < 0

    Where:

    ExcessRi = excess return in period i

    ExcessRmeanBear = mean excess return during bull markets

    N = number of periods

    :param report_id: id of performance report
    :param rolling_window: length of window
    :param benchmark_id: id of benchmark
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of tracking error
    """
    start_date = DataContext.current.start_time.date()
    end_date = DataContext.current.end_time.date()
    rolling_window = _parse_window(rolling_window)
    benchmark_returns = _get_benchmark_return(benchmark_id, start_date, end_date)
    portfolio_returns = _get_daily_pnl(report_id)
    is_bear = benchmark_returns <= 0
    active_return = portfolio_returns - benchmark_returns
    rolling_error_bear = active_return.rolling(window=rolling_window).apply(
        lambda x: np.std(x, ddof=1) if is_bear.loc[x.index[-1]]
        else np.nan,
        raw=False
    )
    return rolling_error_bear


@plot_measure_entity(EntityType.REPORT)
def tracking_error_bull(report_id: str, benchmark_id: str, rolling_window: Union[int, str], *,
                        source: str = None,
                        real_time: bool = False,
                        request_id: Optional[str] = None) -> pd.Series:
    """
    The standard deviation of the excess return relative to the benchmark over the state time frame,
    only counting the periods when benchmark returns were positive.

    Bull Tracking Error = sqrt( sum((ExcessRi - ExcessRmeanBull)^2) / (N-1) ) for all when Benchmark Return > 0

    Where:

    ExcessRi = excess return in period i

    ExcessRmeanBull = mean excess return during bull markets

    N = number of periods

    :param report_id: id of performance report
    :param rolling_window: length of window
    :param benchmark_id: id of benchmark
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of tracking error
    """
    start_date = DataContext.current.start_time.date()
    end_date = DataContext.current.end_time.date()
    rolling_window = _parse_window(rolling_window)
    benchmark_returns = _get_benchmark_return(benchmark_id, start_date, end_date)
    portfolio_pnl = _get_daily_pnl(report_id)
    is_bull = benchmark_returns > 0
    active_return = portfolio_pnl - benchmark_returns
    rolling_error_bull = active_return.rolling(window=rolling_window).apply(
        lambda x: np.std(x, ddof=1) if is_bull.loc[x.index[-1]]
        else np.nan,
        raw=False
    )
    return rolling_error_bull


@plot_measure_entity(EntityType.REPORT)
def portfolio_sharpe_ratio(report_id: str, risk_free_id: str, rolling_window: Union[int, str], *, source: str = None,
                           real_time: bool = False,
                           request_id: Optional[str] = None) -> pd.Series:
    """
    Risk-adjusted measure that calculates the excess return over the risk free rate per unit of volatility

    Sharpe Ratio = (Rp - Rf) / σp

    Where:

    Rp = Portfolio Return

    Rf = Risk-Free Rate

    σp = Standard Deviation of Portfolio Return.

    :param report_id: id of performance report
    :param risk_free_id: id of risk-free asset
    :param rolling_window: length of window
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of sharpe ratio
    """
    start_date = DataContext.current.start_time.date()
    end_date = DataContext.current.end_time.date()
    window = _parse_window(rolling_window)
    portfolio_pnl = _get_daily_pnl(report_id)
    risk_free_rate_daily = (1 + _get_risk_free_rate(risk_free_id, start_date, end_date)) ** (1 / 252) - 1

    daily_excess = (portfolio_pnl - risk_free_rate_daily).dropna()
    rolling_mean = daily_excess.rolling(window).mean()
    rolling_std = daily_excess.rolling(window).std()

    portfolio_sharpe_ratio_series = (rolling_mean / rolling_std) * np.sqrt(252)
    return portfolio_sharpe_ratio_series


@plot_measure_entity(EntityType.REPORT)
def calmar_ratio(report_id: str, rolling_window: Union[int, str], *, source: str = None,
                 real_time: bool = False,
                 request_id: Optional[str] = None) -> pd.Series:
    """
    Risk-adjusted performance metric that divides an investment strategy’s
    annualized rate of return by its maximum drawdown, focusing specifically
    on the strategy’s return per unit of downside risk.

    Calmar Ratio = Annualized Return / Max Drawdown

    :param report_id: id of performance report
    :param rolling_window: size of window to use
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of the calmar ratio
    """
    rolling_window = _parse_window(rolling_window)
    portfolio_pnl = _get_daily_pnl(report_id)
    cumulative_return = (1 + portfolio_pnl) ** np.sqrt(252) - 1
    cumulative_return.index = pd.to_datetime(cumulative_return.index)
    aum_series = aum(report_id)
    max_drawdown_series = max_drawdown(aum_series, rolling_window)
    calmar = cumulative_return / max_drawdown_series.abs()

    return calmar


@plot_measure_entity(EntityType.REPORT)
def sortino_ratio(report_id: str, benchmark_id: str, rolling_window: Union[int, str], *,
                  source: str = None,
                  real_time: bool = False,
                  request_id: Optional[str] = None) -> pd.Series:
    """
    Risk-adjusted measure that calculates the excess return over the benchmark per unit of semi-variance

    Sortino Ratio = (Rp - Rb) / Downside Deviation

    Where:

    Rp = Portfolio Return

    Rb = Benchmark Return

    Downside Deviation = Standard Deviation of negative asset returns.

    :param report_id: id of performance report
    :param benchmark_id: id of benchmark asset, can be security or risk free
    :param rolling_window: size of window to use
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: portfolio estimated pnl
    """
    start_date = DataContext.current.start_time.date()
    end_date = DataContext.current.end_time.date()
    rolling_window = _parse_window(rolling_window)
    portfolio_pnl = _get_daily_pnl(report_id)

    security = SecurityMaster.get_asset(benchmark_id, AssetIdentifier.MARQUEE_ID)

    if isinstance(security, Bond):
        benchmark_pnl = (1 + _get_risk_free_rate(benchmark_id, start_date, end_date)) ** (1 / 252) - 1
    else:
        benchmark_pnl = _get_benchmark_daily_return(benchmark_id, start_date, end_date)

    excess = (portfolio_pnl - benchmark_pnl).dropna()
    sortino = excess.rolling(window=rolling_window).apply(lambda x: _compute_sortino(x), raw=False)
    return sortino


@plot_measure_entity(EntityType.REPORT)
def jensen_alpha(report_id: str, benchmark_id: str, risk_free_id: str, *, source: str = None,
                 real_time: bool = False,
                 request_id: Optional[str] = None) -> pd.Series:
    """
    Risk-Adjusted measure that calculates the actual return of the portfolio
    over and above the return predicted by the CAPM, given the portfolios beta
    and benchmark returns.

    Jensen's Alpha = Rp - [Rf + Beta * (Rb - Rf)]

    Where:

    Rp = Portfolio Return

    Rf = Risk-Free Rate

    Beta = Portfolio Beta

    Rb = Benchmark Return

    :param report_id: id of performance report
    :param benchmark_id: id of benchmark asset
    :param risk_free_id: id of risk-free asset
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of jensen alpha
    """
    start_date = DataContext.current.start_time.date()
    end_date = DataContext.current.end_time.date()
    portfolio_pnl = _get_daily_pnl(report_id)
    benchmark_pnl = _get_benchmark_daily_return(benchmark_id, start_date, end_date)

    risk_free_rate = (1 + _get_risk_free_rate(risk_free_id, start_date, end_date)) ** (1 / 252) - 1

    portfolio_beta_series = beta(portfolio_pnl, benchmark_pnl, prices=False)
    jensen = (portfolio_pnl - (risk_free_rate + portfolio_beta_series * (benchmark_pnl - risk_free_rate)))
    return jensen


@plot_measure_entity(EntityType.REPORT)
def jensen_alpha_bear(report_id: str, benchmark_id: str, risk_free_id: str, *, source: str = None,
                      real_time: bool = False,
                      request_id: Optional[str] = None) -> pd.Series:
    """
    Risk-Adjusted measure that calculates the actual return of the portfolio
    over and above the return predicted by the CAPM, given the portfolios beta
    and benchmark returns. calculated based only on the periods when the benchmark return was negative

    Jensen's Alpha = Rp - [Rf + Beta * (Rb - Rf)]

    Where:

    Rp = Portfolio Return

    Rf = Risk-Free Rate

    Beta = Portfolio Beta

    Rb = Benchmark Return

    :param report_id: id of performance report
    :param benchmark_id: id of benchmark asset
    :param risk_free_id: id of risk-free asset
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of jensen alpha
    """
    start_date = DataContext.current.start_time.date()
    end_date = DataContext.current.end_time.date()
    portfolio_pnl = _get_daily_pnl(report_id)
    benchmark_pnl = _get_benchmark_daily_return(benchmark_id, start_date, end_date)
    risk_free_rate = (1 + _get_risk_free_rate(risk_free_id, start_date, end_date)) ** (1 / 252) - 1
    portfolio_beta_series = beta(portfolio_pnl, benchmark_pnl, prices=False)

    benchmark_pnl = benchmark_pnl[benchmark_pnl < 0]

    jensen = (portfolio_pnl - (risk_free_rate + portfolio_beta_series * (benchmark_pnl - risk_free_rate)))
    return jensen


@plot_measure_entity(EntityType.REPORT)
def jensen_alpha_bull(report_id: str, benchmark_id: str, risk_free_id: str, *, source: str = None,
                      real_time: bool = False,
                      request_id: Optional[str] = None) -> pd.Series:
    """
    Risk-Adjusted measure that calculates the actual return of the portfolio
    over and above the return predicted by the CAPM, given the portfolios beta
    and benchmark returns. calculated based only on the periods when the benchmark return was positive

    Jensen's Alpha = Rp - [Rf + Beta * (Rb - Rf)]

    Where:

    Rp = Portfolio Return

    Rf = Risk-Free Rate

    Beta = Portfolio Beta

    Rb = Benchmark Return

    :param report_id: id of performance report
    :param benchmark_id: id of benchmark asset
    :param risk_free_id: id of risk-free asset
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of jensen alpha
    """
    start_date = DataContext.current.start_time.date()
    end_date = DataContext.current.end_time.date()
    portfolio_pnl = _get_daily_pnl(report_id)
    benchmark_pnl = _get_benchmark_daily_return(benchmark_id, start_date, end_date)
    risk_free_rate = (1 + _get_risk_free_rate(risk_free_id, start_date, end_date)) ** (1 / 252) - 1

    portfolio_beta_series = beta(portfolio_pnl, benchmark_pnl, prices=False)
    benchmark_pnl = benchmark_pnl[benchmark_pnl > 0]
    jensen = (portfolio_pnl - (risk_free_rate + portfolio_beta_series * (benchmark_pnl - risk_free_rate)))
    return jensen


@plot_measure_entity(EntityType.REPORT)
def information_ratio(report_id: str, benchmark_id: str, *, source: str = None,
                      real_time: bool = False,
                      request_id: Optional[str] = None) -> pd.Series:
    """
    Risk-adjusted measure that calculates the excess return over the benchmark,
    per unit of tracking error volatility. It measures the consistency with
    which the portfolio is beating the benchmark.

    Information Ratio = (Rp - Rb) / Tracking Error

    Where:

    Rp = Portfolio Return

    Rb = Benchmark Return

    Tracking Error = Standard Deviation of the difference between the portfolio and benchmark returns.

    :param report_id: id of performance report
    :param benchmark_id: id of benchmark asset
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of the information ratio
    """
    start_date = DataContext.current.start_time.date()
    end_date = DataContext.current.end_time.date()
    portfolio_pnl = _get_daily_pnl(report_id)
    benchmark_pnl = _get_benchmark_daily_return(benchmark_id, start_date, end_date)

    portfolio_excess = portfolio_pnl - benchmark_pnl
    portfolio_std = np.std(portfolio_excess)
    information = portfolio_excess / portfolio_std

    return information


@plot_measure_entity(EntityType.REPORT)
def information_ratio_bear(report_id: str, benchmark_id: str, *, source: str = None,
                           real_time: bool = False,
                           request_id: Optional[str] = None) -> pd.Series:
    """
    Risk-adjusted measure that calculates the excess return over the benchmark,
    per unit of tracking error volatility, counting only the periods when the benchmark return was negative


    Information Ratio = (Rp - Rb) / Tracking Error

    Where:

    Rp = Portfolio Return

    Rb = Benchmark Return

    Tracking Error = Standard Deviation of the difference between the portfolio and benchmark returns.

    :param report_id: id of performance report
    :param benchmark_id: id of benchmark asset
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of the information ratio
    """
    start_date = DataContext.current.start_time.date()
    end_date = DataContext.current.end_time.date()
    portfolio_pnl = _get_daily_pnl(report_id)
    benchmark_pnl = _get_benchmark_daily_return(benchmark_id, start_date, end_date)
    benchmark_pnl = benchmark_pnl[benchmark_pnl < 0]
    portfolio_excess = portfolio_pnl - benchmark_pnl
    portfolio_std = np.std(portfolio_excess)
    information = portfolio_excess / portfolio_std

    return information


@plot_measure_entity(EntityType.REPORT)
def information_ratio_bull(report_id: str, benchmark_id: str, *, source: str = None,
                           real_time: bool = False,
                           request_id: Optional[str] = None) -> pd.Series:
    """
    Risk-adjusted measure that calculates the excess return over the benchmark,
    per unit of tracking error volatility, counting only the periods when the benchmark return was positive.

    Information Ratio = (Rp - Rb) / Tracking Error

    Where:

    Rp = Portfolio Return

    Rb = Benchmark Return

    Tracking Error = Standard Deviation of the difference between the portfolio and benchmark returns.

    :param report_id: id of performance report
    :param benchmark_id: id of benchmark asset
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: portfolio estimated pnl
    """
    start_date = DataContext.current.start_time.date()
    end_date = DataContext.current.end_time.date()
    portfolio_pnl = _get_daily_pnl(report_id)
    benchmark_pnl = _get_benchmark_daily_return(benchmark_id, start_date, end_date)
    benchmark_pnl = benchmark_pnl[benchmark_pnl > 0]
    portfolio_excess = portfolio_pnl - benchmark_pnl
    portfolio_std = np.std(portfolio_excess)
    information = portfolio_excess / portfolio_std

    return information


@plot_measure_entity(EntityType.REPORT)
def modigliani_ratio(report_id: str, benchmark_id: str, risk_free_id: str,
                     rolling_window: Union[int, str], *,
                     source: str = None,
                     real_time: bool = False,
                     request_id: Optional[str] = None) -> pd.Series:
    """
    Modigliani RAP measures how much the portfolio would have returns if it had had
    the same risk as the benchmark. It is a linear transformation of the Sharpe Ratio,
    but the results are expressed in terms of performance.

    M2 = (Sharpe Ratio * σb) + Rf

    Where:

    Sharpe Ratio = Portfolio Sharpe Ratio

    σb = Standard Deviation of the Benchmark Return

    Rf = Risk Free Rate.

    :param report_id: id of performance report
    :param benchmark_id: id of benchmark asset
    :param risk_free_id: id of risk free asset
    :param rolling_window: size of window to use
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of the modigliani ratio
    """
    start_date = DataContext.current.start_time.date()
    end_date = DataContext.current.end_time.date()
    rolling_window = _parse_window(rolling_window)

    benchmark_return = _get_benchmark_daily_return(benchmark_id, start_date, end_date)
    risk_free_rates = (1 + _get_risk_free_rate(risk_free_id, start_date, end_date)) ** (1 / 252) - 1

    sharpe_series = portfolio_sharpe_ratio(report_id, risk_free_id, rolling_window)

    benchmark_excess = benchmark_return - risk_free_rates

    std_benchmark = benchmark_excess.dropna().rolling(window=rolling_window).std()

    modigliani = (sharpe_series * std_benchmark) + risk_free_rates

    return modigliani


@plot_measure_entity(EntityType.REPORT)
def treynor_measure(report_id: str, risk_free_id: str, benchmark_id: str, *, source: str = None,
                    real_time: bool = False,
                    request_id: Optional[str] = None) -> pd.Series:
    """
    Risk-adjusted measure that calculates the excess return over the risk free rate
    per unit of beta relative to the benchmark. This is useful for assessing the excess
    return from each unit of systematic risk.

    Treynor Ratio = (Rp - Rf) / Beta

    Where:

    Rp = Portfolio Return

    Rf = Risk-Free Rate

    Beta = Portfolio Beta

    :param report_id: id of performance report
    :param risk_free_id: id of risk-free asset
    :param benchmark_id: id of benchmark asset
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: portfolio estimated pnl
    """
    start_date = DataContext.current.start_time.date()
    end_date = DataContext.current.end_time.date()
    portfolio_return = _get_daily_pnl(report_id)
    security = SecurityMaster.get_asset(benchmark_id,
                                        AssetIdentifier.MARQUEE_ID)
    spot_data_coordinate = security.get_data_coordinate(DataMeasure.SPOT_PRICE)
    benchmark_pricing = spot_data_coordinate.get_series(start=start_date, end=end_date)
    benchmark_pricing = benchmark_pricing.sort_index()
    portfolio_return = (1 + portfolio_return) ** np.sqrt(252) - 1
    risk_free_rate = _get_risk_free_rate(risk_free_id, start_date, end_date)

    beta_series = beta(aum(report_id), benchmark_pricing)

    treynor_series = (portfolio_return - risk_free_rate) / beta_series
    return treynor_series


@plot_measure_entity(EntityType.REPORT)
def alpha(report_id: str, benchmark_id: str, rolling_window: Union[int, str], *,
          source: str = None,
          real_time: bool = False,
          request_id: Optional[str] = None) -> pd.Series:
    """
    This is the intercept from a regression analysis, where the portfolio's returns are regressed
    against the benchmark's returns. It represents the portion of the portfolio's return that is
    not explained by the benchmark's performance.

    :param report_id: id of performance report
    :param benchmark_id: id of benchmark asset
    :param rolling_window: size of window to use
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of the beta
    """
    start_date = DataContext.current.start_time.date()
    end_date = DataContext.current.end_time.date()
    rolling_window = _parse_window(rolling_window)
    portfolio_pnl = _get_daily_pnl(report_id)

    benchmark_pnl = _get_benchmark_return(benchmark_id, start_date, end_date)

    intercepts = {}
    benchmark_pnl = benchmark_pnl.reindex(portfolio_pnl.index)

    for i in range(rolling_window, len(portfolio_pnl) + 1):
        portfolio_window = portfolio_pnl.iloc[i - rolling_window:i]
        benchmark_window = benchmark_pnl.iloc[i - rolling_window:i]

        slope, intercept, *_ = stats.linregress(portfolio_window, benchmark_window)
        intercepts[portfolio_pnl.index[i - 1]] = intercept

    alpha_series = pd.Series(intercepts)

    return alpha_series


@plot_measure_entity(EntityType.REPORT)
def portfolio_beta(report_id: str, benchmark_id: str, rolling_window: Union[int, str], *,
                   source: str = None,
                   real_time: bool = False,
                   request_id: Optional[str] = None) -> pd.Series:
    """
    Beta defined as the slope of the regression between report pnl and benchmark pnl

    :param report_id: id of performance report
    :param benchmark_id: id of benchmark asset
    :param rolling_window: size of window to use
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of the beta
    """
    start_date = DataContext.current.start_time.date()
    end_date = DataContext.current.end_time.date()
    rolling_window = _parse_window(rolling_window)
    portfolio_return = _get_daily_pnl(report_id)
    benchmark_pnl = _get_benchmark_return(benchmark_id, start_date, end_date)

    return beta(portfolio_return, benchmark_pnl, rolling_window, prices=False)


@plot_measure_entity(EntityType.REPORT)
def portfolio_correlation(report_id: str, benchmark_id: str, rolling_window: Union[int, str],
                          *,
                          source: str = None,
                          real_time: bool = False,
                          request_id: Optional[str] = None) -> pd.Series:
    """
    Correlation coefficient between the portfolio and the benchmark returns over the stated time frame

    :param report_id: id of performance report
    :param benchmark_id: id of benchmark asset
    :param rolling_window: size of window to use
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of the correlation coefficient
    """
    start_date = DataContext.current.start_time.date()
    end_date = DataContext.current.end_time.date()
    rolling_window = _parse_window(rolling_window)
    portfolio_pnl = aum(report_id)
    portfolio_pnl.index = pd.to_datetime(portfolio_pnl.index)

    security = SecurityMaster.get_asset(benchmark_id,
                                        AssetIdentifier.MARQUEE_ID)
    spot_data_coordinate = security.get_data_coordinate(DataMeasure.SPOT_PRICE)
    benchmark_pricing = spot_data_coordinate.get_series(start=start_date, end=end_date)
    benchmark_pricing = benchmark_pricing.sort_index()

    return correlation(portfolio_pnl, benchmark_pricing, rolling_window)


@plot_measure_entity(EntityType.REPORT)
def capture_ratio(report_id: str, benchmark_id: str, rolling_window: Union[int, str], *,
                  source: str = None,
                  real_time: bool = False,
                  request_id: Optional[str] = None) -> pd.Series:
    """
    A measure of how well portfolio is doing relative to the benchmark,
    defined as the ratio of the portfolio return to the benchmark return calculated
    on a daily basis and averaged over the stated time period.

    :param report_id: id of performance report
    :param benchmark_id: id of benchmark asset
    :param rolling_window: size of window to use
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: portfolio capture ratio
    """
    start_date = DataContext.current.start_time.date()
    end_date = DataContext.current.end_time.date()
    rolling_window = _parse_window(rolling_window)
    benchmark_return = _get_benchmark_daily_return(benchmark_id, start_date, end_date)
    portfolio_return = _get_daily_pnl(report_id)

    avg_portfolio = portfolio_return.rolling(window=rolling_window).mean()
    avg_benchmark = benchmark_return.rolling(window=rolling_window).mean()

    capture_ratio_series = (avg_portfolio / avg_benchmark)
    return capture_ratio_series


@plot_measure_entity(EntityType.REPORT)
def r_squared(report_id: str, benchmark_id: str, rolling_window: Union[int, str], *,
              source: str = None,
              real_time: bool = False,
              request_id: Optional[str] = None) -> pd.Series:
    """
    A measure of how well the portfolio's performance correlates with the performance of the benchmark,
    and thus a measure of what portion of its performance may be explained by the performance of the benchmark.

    :param report_id: id of performance report
    :param benchmark_id: id of benchmark
    :param rolling_window: size of rolling window
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: portfolio r squared
    """
    rolling_window = _parse_window(rolling_window)
    r_values = portfolio_correlation(report_id, benchmark_id, rolling_window)
    r_values_squared = r_values ** 2
    return r_values_squared


def _get_benchmark_return(benchmark_id, start_date, end_date):
    security = SecurityMaster.get_asset(benchmark_id,
                                        AssetIdentifier.MARQUEE_ID)
    spot_data_coordinate = security.get_data_coordinate(DataMeasure.SPOT_PRICE)
    benchmark_pricing = spot_data_coordinate.get_series(start=start_date, end=end_date)
    benchmark_pricing = benchmark_pricing.sort_index()

    start_price = benchmark_pricing.iloc[0]

    benchmark_returns = (benchmark_pricing / start_price) - 1
    return benchmark_returns * 100


def _get_benchmark_daily_return(benchmark_id, start_date, end_date):
    security = SecurityMaster.get_asset(benchmark_id,
                                        AssetIdentifier.MARQUEE_ID)
    spot_data_coordinate = security.get_data_coordinate(DataMeasure.SPOT_PRICE)
    benchmark_pricing = spot_data_coordinate.get_series(start=start_date, end=end_date)
    benchmark_pricing = benchmark_pricing.sort_index()
    return benchmark_pricing.pct_change()


def _max_drawdown(series):
    running_max = series.cummax()
    drawdown = (series - running_max) / running_max
    max_drawdown_ts = drawdown.cummin()
    return max_drawdown_ts[-1]


def _compute_sortino(series):
    downside = series[series < 0]
    downside_std = np.sqrt((downside ** 2).mean()) * np.sqrt(252)
    mean_excess = series.mean() * 252
    return mean_excess / downside_std if downside_std != 0 else np.nan


def _rolling_downside_risk(series):
    negative_devs = series[series < series.mean()] - series.mean()
    semi_variance_series = (negative_devs ** 2).mean()
    return np.sqrt(semi_variance_series)


def _rolling_semi_variance(series):
    negative_devs = series[series < 0] - series.mean()
    semi_variance_series = (negative_devs ** 2).mean()
    return np.sqrt(semi_variance_series)


def _get_daily_pnl(report_id: str) -> pd.Series:
    portfolio_return = pnl(report_id)
    portfolio_return.index = pd.to_datetime(portfolio_return.index)
    portfolio_return = portfolio_return / aum(report_id).shift(1)

    return portfolio_return


def _get_risk_free_rate(risk_free_id, start_date, end_date):
    risk_free = SecurityMaster.get_asset(risk_free_id,
                                         AssetIdentifier.MARQUEE_ID)
    risk_free_pricing_data = risk_free.get_data_coordinate('yield')
    risk_free_rate = risk_free_pricing_data.get_series(start=start_date, end=end_date)
    return risk_free_rate.drop_duplicates()


def _replay_historical_factor_moves_on_latest_positions(report_id: str, factors: List[str]) -> \
        Union[pd.Series, pd.DataFrame]:
    start_date = DataContext.current.start_time.date()
    end_date = DataContext.current.end_time.date()
    risk_report = FactorRiskReport.get(report_id)
    risk_model_id = risk_report.get_risk_model_id()

    # Get data in batches of 365 days
    date_range = pd.bdate_range(start_date, end_date)
    batches = np.array_split([d.date() for d in date_range.tolist()], math.ceil(len(date_range.tolist()) / 365))
    data_query_results = []
    query = {"riskModel": risk_model_id}
    if factors:
        query.update({"factor": factors})
    for batch in batches:
        data_query_results += GsDataApi.execute_query(
            'RISK_MODEL_FACTOR',
            DataQuery(
                where=query,
                start_date=batch[0],
                end_date=batch[-1]
            )
        ).get('data', [])

    return_data = pd.DataFrame(data_query_results).pivot(columns="factor", index="date", values="return").sort_index()
    return_data_aggregated = (return_data / 100).apply(geometrically_aggregate)

    latest_report_date = risk_report.latest_end_date
    factor_exposures = risk_report.get_results(
        start_date=latest_report_date,
        end_date=latest_report_date,
        return_format=ReturnFormat.JSON
    )
    factor_exposure_df = pd.DataFrame(factor_exposures).pivot(columns="factor",
                                                              index="date",
                                                              values="exposure").sort_index()

    factor_exposure_df = factor_exposure_df.reindex(columns=return_data_aggregated.columns)
    factor_attributed_pnl_values = return_data_aggregated.values * factor_exposure_df.values
    factor_attributed_pnl = pd.DataFrame(factor_attributed_pnl_values, index=return_data_aggregated.index,
                                         columns=return_data_aggregated.columns)

    return factor_attributed_pnl


def _parse_window(window: Union[int, str]):
    if isinstance(window, int):
        return window
    match = re.fullmatch(r'(\d+)([dwmy])', window.strip().lower())
    if match is None:
        raise MqValueError('Invalid window format, please end with one of: ["d", "w", "m", "y"]')
    value, unit = match.groups()
    value = int(value)
    unit_multipliers = {
        'd': 1,
        'w': 5,
        'm': 22,
        'y': 252
    }
    return value * unit_multipliers[unit]


def _get_factor_data(report_id: str, factor_name: str, query_type: QueryType, unit: Unit = Unit.NOTIONAL) -> pd.Series:
    # Check params
    report = FactorRiskReport.get(report_id)
    if factor_name not in ['Factor', 'Specific', 'Total']:
        if query_type in [QueryType.DAILY_RISK, QueryType.ANNUAL_RISK]:
            raise MqValueError('Please pick a factor name from the following: ["Total", "Factor", "Specific"]')
        model = FactorRiskModel.get(report.get_risk_model_id())
        factor = model.get_factor(factor_name)
        factor_name = factor.name

    # Extract relevant data for each date
    col_name = query_type.value.replace(' ', '')
    col_name = decapitalize(col_name)
    data_type = decapitalize(col_name[6:]) if col_name.startswith('factor') else col_name
    include_total_factor = factor_name != 'Total' and unit == Unit.PERCENT and query_type == QueryType.FACTOR_PNL
    factors_to_query = [factor_name, 'Total'] if include_total_factor else [factor_name]
    factor_data = report.get_results(
        factors=factors_to_query,
        start_date=DataContext.current.start_date,
        end_date=DataContext.current.end_date,
        return_format=ReturnFormat.JSON
    )
    total_data = [d for d in factor_data if d.get(data_type) is not None and d.get('factor') == 'Total']
    factor_data = [d for d in factor_data if d.get(data_type) is not None and d.get('factor') == factor_name]
    if unit == Unit.PERCENT:
        if report.position_source_type != PositionSourceType.Portfolio:
            raise MqValueError('Unit can only be set to percent for portfolio reports')
        pm = PortfolioManager(report.position_source_id)
        tags = dict((tag.name, tag.value) for tag in report.parameters.tags) if report.parameters.tags else None
        performance_report = pm.get_performance_report(tags=tags)
        start_date = dt.datetime.strptime(min([d['date'] for d in factor_data]), '%Y-%m-%d').date()
        end_date = dt.datetime.strptime(max([d['date'] for d in factor_data]), '%Y-%m-%d').date()
        if query_type == QueryType.FACTOR_PNL:
            # Factor Pnl needs to be geometrically aggregated if unit is %
            aum_df = format_aum_for_return_calculation(performance_report, start_date, end_date)
            return get_factor_pnl_percent_for_single_factor(factor_data, total_data, aum_df, start_date)
        else:
            aum = performance_report.get_aum(start_date=start_date, end_date=end_date)
            for data in factor_data:
                if aum.get(data['date']) is None:
                    raise MqValueError('Cannot convert to percent: Missing AUM on some dates in the date range')
            factor_exposures = [{
                'date': d['date'],
                col_name: d[data_type] / aum.get(d['date']) * 100
            } for d in factor_data]
    else:
        factor_exposures = [{'date': d['date'], col_name: d[data_type]} for d in factor_data]

    # Create and return timeseries
    df = pd.DataFrame(factor_exposures)
    if not df.empty:
        df = df.set_index('date')
        df.index = pd.to_datetime(df.index)
    return _extract_series_from_df(df, query_type)


def _return_metrics(one_leg: pd.DataFrame, dates: list, name: str):
    if one_leg.empty:
        return pd.DataFrame(index=dates, data={f'{name}Metrics': [0 for d in dates], "exposure": [0 for d in dates]})
    one_leg = one_leg.groupby(one_leg.index).agg(pnl=('pnl', 'sum'), exposure=('netExposure', 'sum'))

    one_leg['cumulativePnl'] = one_leg['pnl'].cumsum(axis=0)

    one_leg['normalizedExposure'] = (one_leg['exposure'] - one_leg['cumulativePnl'])
    one_leg.iloc[0, one_leg.columns.get_loc('cumulativePnl')] = 0
    one_leg[f'{name}Metrics'] = one_leg['cumulativePnl'] / one_leg['normalizedExposure'] + 1

    one_leg[f'{name}Metrics'] = 1 / one_leg[f'{name}Metrics'] if one_leg['exposure'].iloc[-1] < 0 else one_leg[
        f'{name}Metrics']
    return one_leg
